from __future__ import absolute_import

from contextlib import contextmanager
from collections import defaultdict, deque, namedtuple
import functools
import itertools
import math
import random
import socket
import time

import gevent

from .timer_queue import LOW_RESOLUTION_TIME_SOURCE

class VarzType(object):
  Gauge = 1
  Rate = 2
  AggregateTimer = 3
  Counter = 4
  AverageTimer = 5
  AverageRate = 6

class Source(object):
  __slots__ = "method", "service", "endpoint", "client_id"
  def __init__(self, method=None, service=None, endpoint=None, client_id=None):
    self.method = method
    self.service = service
    self.endpoint = endpoint
    self.client_id = client_id

  def to_tuple(self):
    return (self.method, self.service, self.endpoint, self.client_id)

  def to_dict(self):
    return {
      "method": self.method,
      "service": self.service,
      "endpoint": self.endpoint,
      "client_id": self.client_id,
    }

  def __cmp__(self, other):
    if not isinstance(other, Source):
      return -1
    return cmp(self.to_tuple(), other.to_tuple())

  def __hash__(self):
    return hash((self.method, self.service, self.endpoint, self.client_id))

class VarzMetric(object):
  """An object that can be used to record specific varz."""
  VARZ_TYPE = None

  @staticmethod
  def _Adapt(fn):
    def __Adapt(metric, source, amount=1):
      fn(source, metric, amount)
    return __Adapt

  def __init__(self, metric, source):
    """
    Args:
      metric - The metric name.
      source - An optional source.
    """
    self._metric = metric
    self._source = source
    if self.VARZ_TYPE == VarzType.Gauge:
      self._fn = VarzReceiver.SetVarz
    elif self.VARZ_TYPE in (VarzType.AverageTimer, VarzType.AverageRate):
      self._fn = VarzReceiver.RecordPercentileSample
    else:
      self._fn = VarzReceiver.IncrementVarz

    if source:
      self._fn = functools.partial(self._fn, self._source)
    else:
      self._fn = self._Adapt(self._fn)

  def __call__(self, *args):
    self._fn(self._metric, *args)

  def ForSource(self, source):
    """Specialize this metric for a specific source.

    Args:
      source - The source to specialize for.
    Returns:
      A VarzMetric of the same type.
    """
    if not isinstance(source, Source):
      raise ValueError("Source must be a source object", type(source))
    return type(self)(self._metric, source)

class Gauge(VarzMetric): VARZ_TYPE = VarzType.Gauge
class Rate(VarzMetric): VARZ_TYPE = VarzType.Rate
class AverageRate(VarzMetric): VARZ_TYPE = VarzType.AverageRate
class Counter(Rate): VARZ_TYPE = VarzType.Counter

class VarzTimerBase(VarzMetric):
  """A specialization of VarzMetric that also includes a contextmanager
  to time blocks of code."""
  @contextmanager
  def Measure(self, source=None):
    start_time = time.time()
    yield
    end_time = time.time()
    if source:
      self(source, end_time - start_time)
    else:
      self(end_time - start_time)

class AverageTimer(VarzTimerBase): VARZ_TYPE = VarzType.AverageTimer
class AggregateTimer(VarzTimerBase): VARZ_TYPE = VarzType.AggregateTimer

class VarzMeta(type):
  def __new__(mcs, name, bases, dct):
    base_name = dct['_VARZ_BASE_NAME']
    for metric_suffix, varz_cls in dct['_VARZ'].items():
      metric_name = '%s.%s' % (base_name, metric_suffix)
      VarzReceiver.RegisterMetric(metric_name, varz_cls.VARZ_TYPE)
      varz = varz_cls(metric_name, None)
      dct['_VARZ'][metric_suffix] = varz
      dct[metric_suffix] = varz
    return super(VarzMeta, mcs).__new__(mcs, name, bases, dct)

def VerifySource(source):
  if not isinstance(source, Source):
    raise ValueError("InvalidSource", source)
  return source

class _VarzBase(object):
  """A helper class to create a set of Varz.

  Inheritors should set _VARZ_BASE_NAME.  Once done, the Varz object can be
  instantiated with a source, and then metrics invoked as attributes on the
  object.

  For example:
    class Varz(VarzBase):
      _VARZ_BASE_NAME = 'scales.example.varz'
      _VARZ = {
        'a_counter': Counter,
        'a_gauge': Gauge
      }

    my_varz = Varz(Source(service='my_service'))
    my_varz.a_counter()
    my_varz.a_gauge(5)
  """

  _VARZ = {}
  _VARZ_BASE_NAME = None

  def __init__(self, source):
    source = VerifySource(source)
    for k, v in self._VARZ.items():
      setattr(self, k, v.ForSource(source))

  def __getattr__(self, item):
    return self._VARZ[item]

VarzBase = VarzMeta(
  '_VarzBase',
  (_VarzBase,),
  {
    '_VARZ_BASE_NAME': _VarzBase._VARZ_BASE_NAME,
    '_VARZ': _VarzBase._VARZ
  }
)

class _SampleSet(object):
  __slots__ = ('data', 'i', 'p', 'max_size', 'last_update')

  def __init__(self, max_size, data=None, p=.1):
    data = data or []
    self.data = deque(data, max_size)
    self.i = len(data)
    self.p = p
    self.max_size = max_size
    self.last_update = LOW_RESOLUTION_TIME_SOURCE.now

  def Sample(self, value):
    if self.i < self.max_size:
      self.data.append(value)
      self.last_update = LOW_RESOLUTION_TIME_SOURCE.now
    else:
      j = random.random()
      if j < self.p:
        self.data.append(value)
        self.last_update = LOW_RESOLUTION_TIME_SOURCE.now
    self.i += 1

class VarzReceiver(object):
  """A stub class to receive varz from Scales."""
  VARZ_METRICS = {}
  VARZ_DATA = defaultdict(lambda: defaultdict(int))
  VARZ_PERCENTILES = [.5, .90, .99, .999, .9999]

  _MAX_PERCENTILE_SIZE = 1000

  @staticmethod
  def RegisterMetric(metric, varz_type):
    VarzReceiver.VARZ_METRICS[metric] = varz_type

  @staticmethod
  def IncrementVarz(source, metric, amount=1):
    """Increment (source, metric) by amount"""
    VarzReceiver.VARZ_DATA[metric][VerifySource(source)] += amount

  @staticmethod
  def SetVarz(source, metric, value):
    """Set (source, metric) to value"""
    VarzReceiver.VARZ_DATA[metric][VerifySource(source)] = value

  @classmethod
  def RecordPercentileSample(cls, source, metric, value):
    source = VerifySource(source)
    reservoir = cls.VARZ_DATA[metric][source]
    if reservoir == 0:
      reservoir = _SampleSet(cls._MAX_PERCENTILE_SIZE)
      cls.VARZ_DATA[metric][source] = reservoir
    reservoir.Sample(value)


def DefaultKeySelector(k):
  """A key selector to use for Aggregate."""
  VerifySource(k)
  return k.service, k.client_id


class VarzAggregator(object):
  """An aggregator that rolls metrics up to the service level."""
  MAX_AGG_AGE = 5 * 60

  class _Agg(object):
    __slots__ = 'total', 'count', 'work'

    def __init__(self):
      self.total = 0.0
      self.count = 0
      self.work = 0.0

  @staticmethod
  def CalculatePercentile(values, pct):
    if not values:
      return 0

    k = (len(values) - 1) * pct
    f = math.floor(k)
    c = math.ceil(k)
    if f == c:
      return values[int(k)]
    d0 = values[int(f)] * (c - k)
    d1 = values[int(c)] * (k - f)
    return d0 + d1

  @staticmethod
  def _Downsample(lst, target_size):
    if target_size == 0:
      return
    elif len(lst) < 3 or len(lst) <= target_size:
      for n in lst:
        yield n
    else:
      skip = len(lst) / target_size
      lst = sorted(lst)
      for i, n in enumerate(lst[0:-2]):
        if i % skip == 0:
          yield n
      yield lst[-1]

  @staticmethod
  def Aggregate(varz, metrics, key_selector=None):
    """Aggregate varz

    Args:
      varz - The varz dictionary to aggregate, typically VarzReceiver.VARZ_DATA.
      metrics - The metric metadata dictionary, typically VarzReceiver.VARZ_METRICS
      key_selector - A function to select the key to aggregate by given a
                     (endpoint, service, host, client) tuple.  By default this returns
                     the service.
    Returns:
      A dictionary of metric -> service -> aggregate.
    """
    if not key_selector:
      key_selector = DefaultKeySelector

    agg = defaultdict(dict)
    now = LOW_RESOLUTION_TIME_SOURCE.now
    for metric in varz.keys():
      if metric not in metrics:
        continue
      varz_type = metrics[metric]
      assert isinstance(varz_type, int), varz_type
      metric_agg = agg[metric]
      gevent.sleep(0)
      for source in varz[metric].keys():
        key = key_selector(source)
        data = varz[metric][source]
        if key not in metric_agg:
          metric_agg[key] = VarzAggregator._Agg()
          if isinstance(data, _SampleSet):
            metric_agg[key].work = []

        if isinstance(data, _SampleSet):
          if (now - data.last_update) < VarzAggregator.MAX_AGG_AGE:
            metric_agg[key].work.append(data)
            metric_agg[key].count += 1
        else:
          metric_agg[key].work += data
          metric_agg[key].count += 1
      if varz_type in (VarzType.AggregateTimer, VarzType.Counter,
                       VarzType.Gauge, VarzType.Rate):
        for source_agg in metric_agg.values():
          source_agg.total = source_agg.work
      elif varz_type in (VarzType.AverageTimer, VarzType.AverageRate):
        for source_agg in metric_agg.values():
          if source_agg.count > 0:
            pct_sample = 1.0 / source_agg.count
            values = [VarzAggregator._Downsample(v.data, int(len(v.data) * pct_sample))
                      for v in source_agg.work]
            values = sorted(itertools.chain(*values))
          else:
            values = []
          source_agg.total = [
            VarzAggregator.CalculatePercentile(values, pct)
            for pct in VarzReceiver.VARZ_PERCENTILES
            ]
          if values:
            source_agg.total.insert(0, sum(values) / float(len(values)))
          else:
            source_agg.total.insert(0, 0)
          source_agg.work = None
      else:
        for source_agg in metric_agg.values():
          source_agg.total = float(source_agg.work) / source_agg.count

    return agg

class VarzSocketWrapper(object):
  """A wrapper for Thrift sockets that records various varz about the socket."""
  __slots__ = ('_socket', '_varz', '_is_open')

  class Varz(VarzBase):
    _VARZ_BASE_NAME = 'scales.socket'
    _VARZ = {
      'bytes_recv': Rate,
      'bytes_sent': Rate,
      'num_connections': Counter,
      'tests_failed': Counter,
      'connects': Rate,
      'open_latency': AverageTimer
    }

  def __init__(self, socket, varz_tag):
    self._socket = socket
    self._is_open = self._socket.isOpen()
    self._varz = self.Varz(Source(service=varz_tag, endpoint='%s:%d' % (self.host, self.port)))

  @property
  def host(self):
    return self._socket.host

  @property
  def port(self):
    return self._socket.port

  def isOpen(self):
    return self._socket.isOpen()

  def read(self, sz):
    buff = self._socket.read(sz)
    self._varz.bytes_recv(len(buff))
    return buff

  def recv_into(self, buf, sz):
    return self._socket.handle.recv_into(buf, sz)

  def flush(self):
    pass

  def write(self, buff):
    self._socket.handle.sendall(buff)
    self._varz.bytes_sent(len(buff))

  def open(self):
    with self._varz.open_latency.Measure():
      self._socket.open()

    if self._socket.handle:
      # Disabling nagling (enabling TCP_NODELAY) causes the kernel to not buffer
      # small writes for up to 40 (implementation dependent)ms.  This improves
      # request latency for small requests.
      self._socket.handle.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
    self._is_open = True
    self._varz.connects()
    self._varz.num_connections(1)

  def close(self):
    if self._is_open:
      self._is_open = False
      self._varz.num_connections(-1)
      self._socket.close()

  def readAll(self, sz):
    buff = bytearray(sz)
    view = memoryview(buff)
    have = 0
    while have < sz:
      read_size = sz - have
      chunk_len = self.recv_into(view[have:], read_size)
      have += chunk_len
      if chunk_len == 0:
        raise EOFError()
    self._varz.bytes_recv(sz)
    return buff


class MonoClock(object):
  """A clock whose value is guaranteed to always be increasing.
  Clock skew is compensated.
  """
  def __init__(self):
    self._last = time.time()

  def Sample(self):
    """Return the current time, as reported by time.time(), as long as it has
    increased since the last sample."""
    now = time.time()
    if now - self._last > 0:
      self._last = now
    return self._last


class Ema(object):
  """Calculate an exponential moving average over a window."""
  def __init__(self, window):
    """Args:
      window - The smoothing window, in seconds, to calculate the EMA over.
    """
    self._window = window
    self._time = -1
    self.value = 0.0

  def Update(self, ts, sample):
    """Update the EMA with a new sample
    Args:
      ts - The timestamp, in seconds.
      sample - The sampled value.
    Returns:
      The current EMA after being updated with the sample.
    """
    if self._time == -1:
      self._time = ts
      self.value = float(sample)
    else:
      delta = ts - self._time
      self._time = ts
      window = 0 if self._window == 0 else math.exp(-float(delta) / self._window)
      self.value = (sample * (1-window)) + (self.value * window)
    return self.value



"""Utility methods for manipulating AsyncResults."""

import functools
import sys

import gevent
from gevent import Greenlet
from gevent.event import AsyncResult as g_AsyncResult

class NamedGreenlet(Greenlet):
  def __init__(self, run=None, *args, **kwargs):
    self.name = None
    Greenlet.__init__(self, run, *args, **kwargs)

  @classmethod
  def spawn(cls, name, *args, **kwargs):
    g = cls(*args, **kwargs)
    g.name = name
    g.start()
    return g

  def __repr__(self):
    if self.name:
      return self.name
    else:
      return NamedGreenlet.__repr__(self)

class AsyncResult(g_AsyncResult):
  @staticmethod
  def WhenAll(ars):
    """Returns an AsyncResult representing the state of all AsyncResults passed.

    Args:
      ars - An enumerable of AsyncResults.
    Returns:
      An AsyncResult representing the completion of all ars passed in.  When all
      complete, the AsyncResult will be set to an array of the results of each
      AsyncResult, in the order they were enumerated in.
      If any AsyncResult fails, the return result will fail.
    """

    ret = AsyncResult()
    num_ars = len(ars)
    total = [num_ars]
    results = [None] * num_ars
    def complete(_n, _ar):
      if _ar.exception:
        ret.set_exception(_ar.exception)
      elif not ret.ready():
        total[0] -= 1
        results[_n] = _ar.value
        if total[0] == 0:
          ret.set(results)

    for n, ar in enumerate(ars):
      ar.rawlink(functools.partial(complete, n))
    return ret

  @staticmethod
  def WhenAny(ars):
    """Returns an AsyncResult representing the state of any AsyncResult passed in.
    The return value represents the state of the first AsyncResult to complete, or,
    if all fail, the last to fail.

    Args:
      ars - An enumerable of AsyncResults.
    Returns:
      An AsyncResult representing the state of the first AsyncResult to complete.
      The AsyncResult's value will be set to the value of the first result to
      complete, or, if all fail, the exception thrown by the last to fail.
    """
    ready_ars = [ar for ar in ars if ar.ready()]
    if ready_ars:
      return ready_ars[0]

    ret = AsyncResult()
    total = [len(ars)]
    def complete(_ar):
      total[0] -= 1
      if total[0] == 0 and _ar.exception:
        ret.set_exception(_ar.exception)
      elif not ret.ready() and _ar.successful():
        ret.set(_ar.value)

    for ar in ars:
      ar.rawlink(complete)
    return ret

  @staticmethod
  def FromValue(val):
    if val is None:
      return AsyncResult.Complete()
    else:
      ar = AsyncResult()
      ar.set(val)
      return ar

  @staticmethod
  def Complete():
    """Return an AsyncResult that has completed."""
    return _COMPLETE

  @staticmethod
  def CompleteIn(n):
    """Returns an AsyncResult that completes in <n> seconds

    Args:
      n - The number of seconds to wait before completing.
    """
    ar = AsyncResult()
    def helper():
      ar.set()
    g = Greenlet(helper)
    g.start_later(float(n))
    return ar

  def _SafeLinkHelper(self, fn):
    try:
      self.set(fn())
    except:
      self.set_exception(sys.exc_info()[1])

  def SafeLink(self, fn):
    """Propagate the result of calling fn() on a new greenlet to ar

    Args:
      ar - An AsyncResult.
      fn - The function to execute.
    """
    gevent.spawn(self._SafeLinkHelper, fn)

  def ContinueWith(self, fn, on_hub=True):
    cw_ar = AsyncResult()
    def continue_with_callback(_ar):
      def run():
        try:
          val = fn(_ar)
          cw_ar.set(val)
        except:
          cw_ar.set_exception(sys.exc_info()[1])
      if on_hub:
        run()
      else:
        gevent.spawn(run)
    self.rawlink(continue_with_callback)
    return cw_ar

  def Map(self, fn):
    def mapper(_):
      if self.exception:
        return self
      else:
        return fn(self.value)
    return self.ContinueWith(mapper).Unwrap()

  def _UnwrapHelper(self, target):
    if self.ready():
      # We're ready, propagate the result
      if self.exception:
        target.set_exception(self.exception)
      else:
        if isinstance(self.value, AsyncResult):
          self.value._UnwrapHelper(target)
        else:
          target.set(self.value)
    else:
      self.rawlink(
        functools.partial(AsyncResult._UnwrapHelper, target=target))

  def Unwrap(self):
    unwrapped_ar = AsyncResult()
    self._UnwrapHelper(unwrapped_ar)
    return unwrapped_ar

  @staticmethod
  def TryGet(val):
    if isinstance(val, AsyncResult):
      return val.get()
    else:
      return val

  @staticmethod
  def Run(fn):
    ar = AsyncResult()
    ar.SafeLink(fn)
    return ar

  @staticmethod
  def RunInline(fn):
    ar = AsyncResult()
    ar._SafeLinkHelper(fn)
    return ar

_COMPLETE = AsyncResult()
_COMPLETE.set()

class NoopTimeout(object):
  def start(self): pass
  def cancel(self): pass



from __future__ import absolute_import

from gevent.socket import socket as gsocket
import socket


class ScalesSocket(object):
  def __init__(self, host, port):
    self.host = host
    self.port = port
    self.handle = None

  def isOpen(self):
    return self.handle is not None

  def _resolveAddr(self):
    return socket.getaddrinfo(
      self.host,
      self.port,
      socket.AF_UNSPEC,
      socket.SOCK_STREAM,
      0,
      socket.AI_PASSIVE | socket.AI_ADDRCONFIG)

  def open(self):
    resolved = self._resolveAddr()
    for res in resolved:
      self.handle = gsocket(res[0], res[1])
      try:
        self.handle.connect(res[4])
      except socket.error as e:
        if res is not resolved[-1]:
          continue
        else:
          raise e
      break

  def close(self):
    if self.handle:
      self.handle.close()
      self.handle = None

  def readAll(self, sz):
    buff = b''
    have = 0
    while have < sz:
      chunk = self.read(sz - have)
      have += len(chunk)
      buff += chunk

      if len(chunk) == 0:
        raise EOFError()

    return buff

  def read(self, sz):
    return self.handle.recv(sz)

  def write(self, buff):
    sent = 0
    have = len(buff)
    while sent < have:
      plus = self.handle.send(buff)
      if plus == 0:
        raise EOFError()
      sent += plus
      buff = buff[plus:]



from struct import (
  calcsize,
  pack,
  unpack,
  Struct
)


class Structs(object):
  Byte  = Struct('!B')
  Int16 = Struct('!h')
  Int32 = Struct('!i')
  Int64 = Struct('!q')


class BinaryReader(object):
  def __init__(self, buf):
    self._buf = buf

  def ReadString(self):
    str_len, = Structs.Int16.unpack(self._buf.read(2))
    str_data = self._buf.read(str_len)
    return str_data

  def ReadInt16(self):
    data, = Structs.Int16.unpack(self._buf.read(2))
    return data

  def ReadInt32(self):
    data, = Structs.Int32.unpack(self._buf.read(4))
    return data

  def ReadInt64(self):
    data, = Structs.Int64.unpack(self._buf.read(8))
    return data

  def ReadInt32Array(self):
    num_to_read = self.ReadInt32()
    return unpack('!%di' % num_to_read, self._buf.read(4 * num_to_read))

  def Unpack(self, fmt):
    to_read = calcsize(fmt)
    return unpack(fmt, self._buf.read(to_read))


class BinaryWriter(object):
  def __init__(self, buf):
    self._buf = buf

  def WriteByte(self, val):
    self._buf.write(Structs.Byte.pack(val))

  def WriteInt16(self, val):
    self._buf.write(Structs.Int16.pack(val))

  def WriteInt32(self, val):
    self._buf.write(Structs.Int32.pack(val))

  def WriteInt64(self, val):
    self._buf.write(Structs.Int64.pack(val))

  def WriteString(self, val):
    self._buf.write(Structs.Int16.pack(len(val)))
    self._buf.write(val)

  def WriteBinary(self, val):
    self._buf.write(Structs.Int32.pack(len(val)))
    self._buf.write(val)

  def WriteRaw(self, val):
    self._buf.write(val)

  def WriteStruct(self, s, *vals):
    self._buf.write(s.pack(*vals))

  def Pack(self, fmt, *args):
    self._buf.write(pack(fmt, *args))



"""Aperture Load Balancer.
Based on work from finagle's aperture load balancer.
See https://github.com/twitter/finagle/blob/master/finagle-core/src/main/scala/com/twitter/finagle/loadbalancer/Aperture.scala

The aperture balancer attempts to keep the average load going into the underlying
server set between a load band (by default .5 <= load <= 2.

Load is determined via an ema of load over a smoothing window (5 seconds).
The load average is essentially the average number of concurrent requests each
node in the balancer is handling.
"""

import random

from .heap import HeapBalancerSink
from ..asynchronous import AsyncResult
from ..constants import (ChannelState, SinkProperties, SinkRole)
from ..sink import SinkProvider
from ..timer_queue import LOW_RESOLUTION_TIMER_QUEUE, LOW_RESOLUTION_TIME_SOURCE
from ..varz import (
  Ema,
  Gauge,
  MonoClock,
  Source,
  VarzBase
)


class ApertureBalancerSink(HeapBalancerSink):
  """A load balancer that keeps an aperture adjusted by a load average."""

  class ApertureVarz(VarzBase):
    """
    idle - The number of nodes idle in the pool (not in the aperture)
    active - The number of nodes active in the pool (in the aperture)
    load_average - The most recently calculated load average.
    """
    _VARZ_BASE_NAME = 'scales.loadbalancer.Aperture'
    _VARZ = {
      'idle': Gauge,
      'active': Gauge,
      'load_average': Gauge
    }

  def __init__(self, next_provider, sink_properties, global_properties):
    self._idle_endpoints = set()
    self._total = 0
    self._ema = Ema(5)
    self._time = MonoClock()
    self._min_size = sink_properties.min_size
    self._max_size = sink_properties.max_size
    self._min_load = sink_properties.min_load
    self._max_load = sink_properties.max_load
    self._jitter_min = sink_properties.jitter_min_sec
    self._jitter_max = sink_properties.jitter_max_sec
    service_name = global_properties[SinkProperties.Label]
    self.__varz = self.ApertureVarz(Source(service=service_name))
    self._pending_endpoints = set()
    super(ApertureBalancerSink, self).__init__(next_provider, sink_properties, global_properties)
    if self._jitter_min > 0:
      self._ScheduleNextJitter()

  def _UpdateSizeVarz(self):
    """Update active and idle varz"""
    self.__varz.active(self._size)
    self.__varz.idle(len(self._idle_endpoints))

  def _AddSink(self, endpoint, sink_factory):
    """Invoked when a node is added to the underlying server set.

    If the number of healthy nodes is < the minimum aperture size, the node
    will be added to the aperture, otherwise it will be added to the idle channel
    list.

    Args:
      endpoint - The endpoint being added to the server set.
      sink_factory - A callable used to create a sink for the endpoint.
    """
    num_healthy = len([c for c in self._heap[1:] if c.channel.is_open])
    if num_healthy < self._min_size:
      super(ApertureBalancerSink, self)._AddSink(endpoint, sink_factory)
    else:
      self._idle_endpoints.add(endpoint)
    self._UpdateSizeVarz()

  def _RemoveSink(self, endpoint):
    """Invoked when a node is removed from the underlying server set.

    If the node is currently active, it is removed from the aperture and replaced
    by an idle node (if one is available).  Otherwise, it is simply discarded.

    Args:
      endpoint - The endpoint being removed from the server set.
    """
    removed = super(ApertureBalancerSink, self)._RemoveSink(endpoint)
    if removed:
      self._TryExpandAperture()
    if endpoint in self._idle_endpoints:
      self._idle_endpoints.discard(endpoint)
    self._UpdateSizeVarz()

  def _TryExpandAperture(self, leave_pending=False):
    """Attempt to expand the aperture.  By calling this it's assumed the aperture
    needs to be expanded.

    The aperture can be expanded if there are idle sinks available.
    """
    endpoints = list(self._idle_endpoints)
    added_node = None
    new_endpoint = None
    if endpoints:
      new_endpoint = random.choice(endpoints)
      self._idle_endpoints.discard(new_endpoint)
      self._log.debug('Expanding aperture to include %s.' % str(new_endpoint))
      new_sink = self._servers[new_endpoint]
      self._pending_endpoints.add(new_endpoint)
      added_node = super(ApertureBalancerSink, self)._AddSink(new_endpoint, new_sink)

    self._UpdateSizeVarz()
    if added_node:
      if not leave_pending:
        added_node.ContinueWith(
            lambda ar: self._pending_endpoints.discard(new_endpoint))
      return added_node, new_endpoint
    else:
      return AsyncResult.Complete(), None

  def _ContractAperture(self, force=False):
    """Attempt to contract the aperture.  By calling this it's assume the aperture
    needs to be contracted.

    The aperture can be contracted if it's current size is larger than the
    min size.
    """
    if self._pending_endpoints and not force:
      return

    num_healthy = len([c for c in self._heap[1:] if c.channel.is_open])
    if num_healthy > self._min_size:
      least_loaded_endpoint = None
      # Scan the heap for any closed endpoints.
      for n in self._heap[1:]:
        if n.channel.is_closed and n.endpoint not in self._pending_endpoints:
          least_loaded_endpoint = n.endpoint
          break
      if not least_loaded_endpoint:
        # Scan the heap for the least-loaded node.  This isn't exactly in-order,
        # but "close enough"
        for n in self._heap[1:]:
          if n.endpoint not in self._pending_endpoints:
            least_loaded_endpoint = n.endpoint
            break

      if least_loaded_endpoint:
        self._idle_endpoints.add(least_loaded_endpoint)
        super(ApertureBalancerSink, self)._RemoveSink(least_loaded_endpoint)
        self._log.debug('Contracting aperture to remove %s' % str(least_loaded_endpoint))
        self._UpdateSizeVarz()

  def _OnNodeDown(self, node):
    """Invoked by the base class when a node is marked down.
    In this case, if the downed node is currently in the aperture, we want to
    remove if, and then attempt to adjust the aperture.
    """
    if node.channel.state != ChannelState.Idle:
      ar, _ = self._TryExpandAperture()
      return ar
    else:
      return AsyncResult.Complete()

  def _OnGet(self, node):
    """Invoked by the parent class when a node has been retrieved from the pool
    and is about to be used.
    Increases the load average of the pool, and adjust the aperture if needed.
    """
    self._AdjustAperture(1)

  def _OnPut(self, node):
    """Invoked by the parent class when a node is being returned to the pool.
    Decreases the load average and adjust the aperture if needed.
    """
    self._AdjustAperture(-1)

  def _ScheduleNextJitter(self):
    """Schedule the aperture to jitter in a random amount of time between
    _jitter_min and _jitter_max.
    """
    next_jitter = random.randint(self._jitter_min, self._jitter_max)
    now = LOW_RESOLUTION_TIME_SOURCE.now
    self._next_jitter = LOW_RESOLUTION_TIMER_QUEUE.Schedule(
        now + next_jitter, self._Jitter)

  def _Jitter(self):
    """Attempt to expand the aperture by one node, and if successful,
    contract it by a node (excluding the one that was just added).  This is
    done asynchronously.
    """
    try:
      ar, endpoint = self._TryExpandAperture(True)
      if endpoint:
        try:
          ar.wait()
          if not ar.exception:
            self._ContractAperture(True)
        finally:
          self._pending_endpoints.discard(endpoint)
    finally:
      self._ScheduleNextJitter()

  def _AdjustAperture(self, amount):
    """Adjusts the load average of the pool, and adjusts the aperture size
    if required by the new load average.

    Args:
      amount - The amount to change the load by.  May be +/-1
    """
    self._total += amount
    avg = self._ema.Update(self._time.Sample(), self._total)
    aperture_size = self._size
    if aperture_size == 0:
      # Essentially infinite load.
      aperture_load = self._max_load
    else:
      aperture_load = avg / aperture_size
      self.__varz.load_average(aperture_load)
    if (aperture_load >= self._max_load
        and self._idle_endpoints
        and aperture_size < self._max_size):
      self._TryExpandAperture()
    elif aperture_load <= self._min_load and aperture_size > self._min_size:
      self._ContractAperture()


ApertureBalancerSink.Builder = SinkProvider(
  ApertureBalancerSink,
  SinkRole.LoadBalancer,
  smoothing_window = 5,
  min_size = 1,
  max_size = 2**31,
  min_load = 0.5,
  max_load = 2.0,
  server_set_provider = None,
  jitter_min_sec = 120,
  jitter_max_sec = 240)



from struct import pack, Struct
from collections import namedtuple
from ..binary import (
  BinaryReader,
  BinaryWriter
)
from ..constants import TransportHeaders
from ..message import MethodReturnMessage
from ..constants import MessageProperties

import zlib

class MessageHelper(object):
  @staticmethod
  def GetPutArgs(msg):
    def _get_put_args(topic, payloads=[], acks=1):
      return topic, payloads, acks
    return _get_put_args(*msg.args, **msg.kwargs)

class MessageType(object):
  MetadataRequest = 3
  ProduceRequest = 0

class ErrorCode(object):
  NoError = 0
  Unknown = -1
  OffsetOutOfRange = 1
  InvalidMessage = 2
  UnknownTopicOrPartition = 3
  InvalidMessageSize = 4
  LeaderNotAvailable = 5
  NotLeaderForPartition = 6
  RequestTimedOut = 7
  BrokerNotAvailable = 8
  ReplicaNotAvailable = 9
  MessageSizeTooLarge = 10
  StaleControllerEpochCode = 11
  OffsetMetadataTooLargeCode = 12
  OffsetsLoadInProgressCode = 14
  ConsumerCoordinatorNotAvailableCode = 15
  NotCoordinatorForConsumerCode = 16

  ReloadMetadataCodes = (
    UnknownTopicOrPartition,
    NotLeaderForPartition,
    BrokerNotAvailable
  )

  _ErrorCodeLookup = {
   -1: 'Unknown',
    0: 'NoError',
    1: 'OffsetOutOfRange',
    2: 'InvalidMessage',
    3: 'UnknownTopicOrPartition',
    4: 'InvalidMessageSize',
    5: 'LeaderNotAvailable',
    6: 'NotLeaderForPartition',
    7: 'RequestTimedOut',
    8: 'BrokerNotAvailable',
    9: 'ReplicaNotAvailable',
    10: 'MessageSizeTooLarge',
    11: 'StaleControllerEpochCode',
    12: 'OffsetMetadataTooLargeCode',
    14: 'OffsetsLoadInProgressCode',
    15: 'ConsumerCoordinatorNotAvailableCode',
    16: 'NotCoordinatorForConsumerCode',
  }
  @staticmethod
  def MessageForCode(code):
    return ErrorCode._ErrorCodeLookup[code]


class KafkaError(Exception):
  def __init__(self, message, error_code):
    self.message = message
    self.error_code = error_code
    super(KafkaError, self).__init__(message)

class NoBrokerForTopicException(Exception): pass

MetadataResponse = namedtuple('MetadataResponse', 'brokers topics')
BrokerMetadata = namedtuple('BrokerMetadata', 'nodeId host port')
PartitionMetadata = namedtuple('PartitionMetadata', 'topic_name partition_id leader replicas isr')
ProduceResponse = namedtuple('ProduceResponse', 'topic partition error offset')

class KafkaProtocol(object):
  MSG_STRUCT = Struct('!BBii')
  MSG_HEADER = Struct('!qiI')
  PRODUCE_HEADER = Struct('!hii')

  def DeserializeMessage(self, buf, msg_type):
    # Skip the correlationId
    buf.read(4)

    if msg_type == MessageType.MetadataRequest:
      return self._DeserializeMetadataResponse(buf)
    elif msg_type == MessageType.ProduceRequest:
      return self._DeserializeProduceResponse(buf)

  def SerializeMessage(self, msg, buf, headers):
    if msg.method == '__metadata':
      self._SerializeMetadataRequest(msg, buf, headers)
      return MessageType.MetadataRequest
    if msg.method == 'Put':
      self._SerializeProduceRequest(msg, buf, headers)
      return MessageType.ProduceRequest
    else:
      raise NotImplementedError()

  def _SerializeMetadataRequest(self, msg, buf, headers):
    headers[TransportHeaders.MessageType] = MessageType.MetadataRequest
    buf.write(pack('!i', len(msg.args or [])))
    for topic in msg.args:
      buf.write(pack('!h'), len(topic))
      buf.write(topic)

  def _DeserializeMetadataResponse(self, buf):
    reader = BinaryReader(buf)

    num_brokers = reader.ReadInt32()
    brokers = {}
    for n in range(num_brokers):
      nodeId = reader.ReadInt32()
      host = reader.ReadString()
      port = reader.ReadInt32()
      brokers[nodeId] = BrokerMetadata(nodeId, host, port)

    num_topics = reader.ReadInt32()
    topic_metadata = {}
    for t in range(num_topics):
      topic_error_code = reader.ReadInt16() #Unused
      topic_name = reader.ReadString()
      num_partitions = reader.ReadInt32()

      partition_data = {}
      for p in range(num_partitions):
        error_code, partition_id, leader = reader.Unpack('!hii')
        replicas = reader.ReadInt32Array()
        isr = reader.ReadInt32Array()
        partition_data[partition_id] = PartitionMetadata(
          topic_name, partition_id, leader, replicas, isr)

      topic_metadata[topic_name] = partition_data
    return MethodReturnMessage(MetadataResponse(brokers, topic_metadata))

  def _GetMessageHeader(self, payload):
    return self.MSG_STRUCT.pack(0, 0, -1, len(payload))

  def _SerializeProduceRequest(self, msg, buf, headers):
    headers[TransportHeaders.MessageType] = MessageType.ProduceRequest
    topic, payloads, acks = MessageHelper.GetPutArgs(msg)
    endpoint = msg.properties[MessageProperties.Endpoint]

    writer = BinaryWriter(buf)

    # Header, 1000ms timeout, 1 message set
    writer.WriteStruct(self.PRODUCE_HEADER, acks, 1000, 1)

    writer.WriteString(topic)
    writer.WriteInt32(1) # 1 payload
    writer.WriteInt32(endpoint.partition_id)

    # [Offset (8) + HeaderLen (4) + CRC (4) + payload (P) + header(10)] x N messages
    msg_set_len = sum([8 + 4 + 4 + len(p) + 10 for p in payloads])
    writer.WriteInt32(msg_set_len)

    for p in payloads:
      header = self._GetMessageHeader(p)
      # Calculate the CRC
      crc = zlib.crc32(header)
      crc = zlib.crc32(p, crc)

      # Write the header
      writer.WriteStruct(self.MSG_HEADER, 0, len(header) + len(p) + 4, crc & 0xffffffff)
      # Write the message data
      writer.WriteRaw(header)
      writer.WriteRaw(p)

  def _DeserializeProduceResponse(self, buf):
    reader = BinaryReader(buf)
    num_topics = reader.ReadInt32()
    responses = []
    for t in range(num_topics):
      topic = reader.ReadString()
      num_partitions = reader.ReadInt32()
      for p in range(num_partitions):
        partition = reader.ReadInt32()
        error_code = reader.ReadInt16()
        offset = reader.ReadInt64()
        responses.append(ProduceResponse(topic, partition, error_code, offset))
    return MethodReturnMessage(responses)



from collections import namedtuple
from struct import (pack, unpack)
import time

from ..compat import BytesIO
from ..dispatch import MessageDispatcher
from ..loadbalancer.serverset import StaticServerSetProvider
from ..loadbalancer.zookeeper import Member
from ..message import (
  MethodCallMessage,
  MethodReturnMessage
)
from ..mux.sink import MuxSocketTransportSink
from ..sink import (
  ClientMessageSink,
  SinkProperties,
  SinkProvider,
  SinkRole,
)
from ..thrift.sink import SocketTransportSinkProvider

from .protocol import (
  BrokerMetadata,
  ErrorCode,
  KafkaError,
  KafkaProtocol,
  MessageHelper,
  NoBrokerForTopicException,
)

KafkaEndpoint = namedtuple("KafkaEndpoint", "host port partition_id")
KafkaEndpoint.__str__ = lambda k: '%s:%d' % (k.host, k.port)

class TopicState(object):
  def __init__(self, topic_name, lb, brokers, open_ar):
    self.topic_name = topic_name
    self.lb = lb
    self.brokers = brokers
    self.last_refreshed = time.time()
    self.open_ar = open_ar


class KafkaTransportSink(MuxSocketTransportSink):
  CLIENT_ID = 'scales'

  def _CheckInitialConnection(self):
    pass

  def _OnTimeout(self, tag):
    pass

  def _BuildHeader(self, tag, msg_type, data_len):
    header = pack(
      '!ihhih%ds' % len(self.CLIENT_ID),
      2 + 2 + 4 + 2 + len(self.CLIENT_ID) + data_len,
      msg_type,
      0,
      tag,
      len(self.CLIENT_ID),
      self.CLIENT_ID
    )
    return header

  def _ProcessReply(self, stream):
    tag, = unpack('!i', stream.read(4))
    self._ProcessTaggedReply(tag, stream)


class KafkaRouterSink(ClientMessageSink):
  _RETRY_KEY = "__KafkaRetries"
  _BOOTSTRAP_SUFFIX = "-bootstrap"

  def __init__(self, next_provider, sink_properties, global_properties):
    self._metadata = None
    self._topics = {}
    self._next_provider = next_provider
    self._global_properties = global_properties
    self._label = global_properties['label']
    self._refresh_rate = sink_properties.refresh_rate
    self._refresh_ar = None
    self._bootstrap_servers = set()
    self._bootstrap_lb = self._CreateBootstrapLoadBalancer(
      sink_properties.server_set_provider)
    super(KafkaRouterSink, self).__init__()

  def _CreateBootstrapLoadBalancer(self, broker_provider):
    """Create a load balancer used to bootstrap the cluster.

    Args:
      broker_provider: A ServerSetProvider that contains at least one kafka
                       broker.
    Returns:
      A load balancer that can be used to bootstrap the kafka client.
    """
    brokers = []
    broker_provider.Initialize(None, None)
    for bs in broker_provider.GetServers():
      ep = bs.service_endpoint
      brokers.append((-1, BrokerMetadata(-1, ep.host, ep.port)))

    self._bootstrap_servers = set([(b.host, b.port) for _, b in brokers])
    return self._CreateBrokerLoadBalancer(brokers, self._BOOTSTRAP_SUFFIX)

  def _RefreshMetadataAsync(self):
    """Asynchronously refreshes the kafka cluster metadata.

    Returns:
      An AsyncResult representing the refresh operation.
    """
    if not self._refresh_ar:
      self._refresh_ar = self._bootstrap_lb.Open().ContinueWith(
      lambda ar: MessageDispatcher.StaticDispatchMessage(
        self._bootstrap_lb,
        None,
        0,
        0,
        MethodCallMessage(None, '__metadata', [], {}))
      ).Unwrap()
    return self._refresh_ar

  def _CreateBrokerLoadBalancer(self, brokers, service_suffix):
    """Create a load balancer for a set of brokers.

    Args:
      brokers - A list of (partition_id, BrokerMetadata) tuples.
      service_suffix - An optional suffix to add the to generated service.

    Returns:
      A new load balancer sink populated with the brokers passed in.
    """
    members = [
      Member(b.nodeId, KafkaEndpoint(b.host, b.port, partition_id), {}, b.nodeId, 'ALIVE')
      for partition_id, b in brokers
    ]
    ssp = StaticServerSetProvider(members)
    balancer = self._next_provider.Clone(server_set_provider=ssp)
    return balancer.CreateSink({ SinkProperties.Label: self._label + (service_suffix or '') })

  def _RebuildTopicLoadBalancers(self, broker_metadata, requested_topic_name):
    """Creates a new set of load balancers for all topics.  Each load balancer
    consists of the leaders for all partitions in a topic.

    Args:
      broker_metadata - A MetadataResponse object.
      topic_name - The topic to refresh
    """
    brokers = broker_metadata.brokers
    new_topics = {}
    old_lbs_to_close = []
    seen_new_topics = set()

    for topic_name, t in broker_metadata.topics.items():
      partition_leaders = set((p.partition_id, p.leader)
                              for p in t.values() if p.leader != -1)
      topic_brokers = set((partition_id, brokers[broker])
                       for partition_id, broker in partition_leaders)
      topic_state = self._topics.get(topic_name)
      if topic_state and topic_state.brokers == topic_brokers:
        topic_state.last_refreshed = time.time()
      elif topic_name == requested_topic_name:
        if topic_state:
          old_lbs_to_close.append(topic_state.lb)
        lb = self._CreateBrokerLoadBalancer(topic_brokers, None)
        open_ar = lb.Open()
        topic_state = TopicState(topic_name, lb, topic_brokers, open_ar)
      else:
        seen_new_topics.add(topic_name)

      if topic_state:
        new_topics[topic_name] = topic_state

    self._topics = new_topics
    for lb in old_lbs_to_close:
      lb.Close()

    new_bootstrap_keys = set((b.host, b.port) for b in brokers.values())
    if new_bootstrap_keys != self._bootstrap_servers:
      new_bootstrap = self._CreateBrokerLoadBalancer(brokers.items(), self._BOOTSTRAP_SUFFIX)
      old_bootstrap = self._bootstrap_lb
      self._bootstrap_lb = new_bootstrap
      old_bootstrap.Close()

  def _GetTopic(self, topic):
    return self._topics.get(topic)

  def _RefreshBrokersAndRetry(self, topic_name, sink_stack, msg):
    topic_info = self._GetTopic(topic_name)
    if topic_info and (
        time.time() - topic_info.last_refreshed < self._refresh_rate):
      # We've already tried to refresh within the refresh rate, just send it
      self._AsyncProcessRequestToTopic(sink_stack, msg, topic_info)
      return

    refresh_ar = self._RefreshMetadataAsync()
    def _on_refresh_complete(ar):
      self._refresh_ar = None
      if ar.exception:
        sink_stack.AsyncProcessResponseMessage(
          MethodReturnMessage(error=ar.exception))
      else:
        try:
          # Rebuild
          self._RebuildTopicLoadBalancers(ar.value, topic_name)
          # Retry
          topic, _, _ = MessageHelper.GetPutArgs(msg)
          topic_info = self._GetTopic(topic)
          if not topic_info:
            sink_stack.AsyncProcessResponseMessage(
              MethodReturnMessage(error=NoBrokerForTopicException("No broker for topic %s" % topic))
            )
          else:
            # We're in the gevent hub here, move to a new greenlet to do the retry.
            self._AsyncProcessRequestToTopic(sink_stack, msg, topic_info)
        except Exception as ex:
          sink_stack.AsyncProcessResponseMessage(MethodReturnMessage(error=ex))
    refresh_ar.ContinueWith(_on_refresh_complete, False)

  def _AsyncProcessRequestToTopic(self, sink_stack, msg, topic_info):
    def _send_msg(_):
      sink_stack.Push(self, (msg, topic_info.topic_name))
      topic_info.lb.AsyncProcessRequest(sink_stack, msg, None, {})
    if not topic_info.open_ar:
      topic_info.open_ar = topic_info.lb.Open()
    topic_info.open_ar.ContinueWith(_send_msg, False)

  def AsyncProcessRequest(self, sink_stack, msg, stream, headers):
    topic, _, _ = MessageHelper.GetPutArgs(msg)
    topic_info = self._GetTopic(topic)
    if not topic_info:
      self._RefreshBrokersAndRetry(topic, sink_stack, msg)
    else:
      self._AsyncProcessRequestToTopic(sink_stack, msg, topic_info)

  def AsyncProcessResponse(self, sink_stack, context, stream, msg):
    req_msg, topic_name = context
    num_retries = req_msg.properties.get(self._RETRY_KEY, 0)
    if msg.return_value:
      if len(msg.return_value) != 1:
        sink_stack.AsyncProcessResponseMessage(
          MethodReturnMessage(error=Exception('Invalid response from kafka: unexpected return value')))
      else:
        err_code = msg.return_value[0].error
        if err_code == ErrorCode.NoError:
          sink_stack.AsyncProcessResponseMessage(msg)
        elif err_code in ErrorCode.ReloadMetadataCodes and num_retries == 0:
          # Refresh metadata and retry
          req_msg.properties[self._RETRY_KEY] = 1
          self._RefreshBrokersAndRetry(topic_name, sink_stack, req_msg)
        else:
          err_msg = MethodReturnMessage(error=KafkaError(
              'Kafka broker returned error %d' % err_code, err_code))
          sink_stack.AsyncProcessResponseMessage(err_msg)
    else:
      # An exception occured, retry once
      self._RefreshBrokersAndRetry(topic_name, sink_stack, req_msg)


class KafkaSerializerSink(ClientMessageSink):
  def __init__(self, next_provider, sink_properties, global_properties):
    super(KafkaSerializerSink, self).__init__()
    self._serializer = KafkaProtocol()
    self.next_sink = next_provider.CreateSink(global_properties)

  def AsyncProcessRequest(self, sink_stack, msg, stream, headers):
    buf = BytesIO()
    headers = {}

    try:
      ctx = self._serializer.SerializeMessage(msg, buf, headers)
    except Exception as ex:
      msg = MethodReturnMessage(error=ex)
      sink_stack.AsyncProcessResponseMessage(msg)
      return

    sink_stack.Push(self, ctx)
    self.next_sink.AsyncProcessRequest(sink_stack, msg, buf, headers)

  def AsyncProcessResponse(self, sink_stack, context, stream, msg):
    if msg:
      sink_stack.AsyncProcessResponseMessage(msg)
    else:
      try:
        msg = self._serializer.DeserializeMessage(stream, context)
      except Exception as ex:
        msg = MethodReturnMessage(error=ex)
      sink_stack.AsyncProcessResponseMessage(msg)


KafkaRouterSink.Builder = SinkProvider(
  KafkaRouterSink,
  SinkRole.LoadBalancer,
  server_set_provider = None,
  refresh_rate = 10
)
KafkaSerializerSink.Builder = SinkProvider(KafkaSerializerSink)
KafkaTransportSink.Builder = SocketTransportSinkProvider(KafkaTransportSink)



class Enum(object): pass

# States <= Busy are considered open and healthy,
# > Busy are closed and unhealthy.
class ChannelState(object):
  Idle = 1
  Open = 2
  Busy = 3
  Closed = 4


class Int(object):
  MaxValue = 2147483647
  MinValue = -2147483648

class SinkProperties(object):
  Endpoint = 'endpoint'
  ServiceInterface = 'service_iface'
  Label = 'label'

class MessageProperties(object):
  Endpoint = '__Endpoint'

class SinkRole(object):
  Transport = 'transport'
  Pool = 'pool'
  LoadBalancer = 'loadbalancer'
  Formatter = 'formatter'

class TransportHeaders(object):
  MessageType = '__MessageType'



from ..resurrector import ResurrectorSink
from ..loadbalancer import ApertureBalancerSink
from ..pool import WatermarkPoolSink
from .sink import (
  SocketTransportSink,
  ThriftSerializerSink,
)

from ..core import Scales


class Thrift(object):
  """A builder for Thrift clients."""

  @staticmethod
  def NewBuilder(Iface):
    return Scales.NewBuilder(Iface)\
      .WithSink(ThriftSerializerSink.Builder())\
      .WithSink(ApertureBalancerSink.Builder())\
      .WithSink(ResurrectorSink.Builder())\
      .WithSink(WatermarkPoolSink.Builder())\
      .WithSink(SocketTransportSink.Builder())

  @staticmethod
  def NewClient(Iface, uri, timeout=60):
    return Thrift.NewBuilder(Iface)\
      .SetUri(uri)\
      .SetTimeout(timeout)\
      .Build()



"""An abstract base class for load balancer.
Load balancers track an underlying server set, and return a sink to service a
request.
"""
from abc import abstractmethod
import functools
import logging
import random

import gevent
from gevent.event import Event

from ..asynchronous import AsyncResult
from ..constants import (
  ChannelState,
  SinkProperties,
  SinkRole
)
from ..message import Deadline
from ..sink import ClientMessageSink

ROOT_LOG = logging.getLogger("scales.loadbalancer")

class NoMembersError(Exception): pass

class LoadBalancerSink(ClientMessageSink):
  """Base class for all load balancer sinks."""
  Role = SinkRole.LoadBalancer

  def __init__(self, next_provider, sink_properties, global_properties):
    self._properties = global_properties
    service_name = global_properties[SinkProperties.Label]
    server_set_provider = sink_properties.server_set_provider
    log_name = self.__class__.__name__.replace('ChannelSink', '')
    self._log = ROOT_LOG.getChild('%s.[%s]' % (log_name, service_name))
    self.__init_done = Event()
    self.__open_ar = None
    self.__open_greenlet = None
    self._server_set_provider = server_set_provider
    self._endpoint_name = server_set_provider.endpoint_name
    self._next_sink_provider = next_provider
    self._state = ChannelState.Idle
    self._servers = {}
    super(LoadBalancerSink, self).__init__()

  @property
  def state(self):
    return self._state

  def Open(self):
    if self.__open_ar:
      return self.__open_ar

    self.__open_ar = AsyncResult()
    self.__open_greenlet = gevent.spawn(self._OpenImpl)
    return self.__open_ar

  def _OpenImpl(self):
    while self._state != ChannelState.Closed:
      try:
        self._server_set_provider.Initialize(
            self.__OnServerSetJoin,
            self.__OnServerSetLeave)
        server_set = self._server_set_provider.GetServers()
      except gevent.GreenletExit:
        return
      except:
        self._log.exception("Unable to initialize serverset, retrying in 5 seconds.")
        gevent.sleep(5)
        continue

      random.shuffle(server_set)
      self._servers = {}
      [self.__AddServer(m) for m in server_set]
      self.__init_done.set()
      self._OpenInitialChannels()
      self._open_greenlet = None
      self._state = ChannelState.Open
      return True

  @abstractmethod
  def _OpenInitialChannels(self):
    """To be overriden by subclasses.  Called after the ServerSet is initialized
    and the initial set of servers has been loaded.
    """
    pass

  def _OnOpenComplete(self):
    """To be called by subclasses when they've completed (successfully or not)
    opening their sinks.
    """
    self.__open_ar.set(True)

  def WaitForOpenComplete(self, timeout=None):
    self.__open_ar.wait(timeout)

  def Close(self):
    self._server_set_provider.Close()
    self._state = ChannelState.Closed
    self.__init_done.clear()
    if self.__open_greenlet:
      self.__open_greenlet.kill(block=False)
      self.__open_greenlet = None
    self.__open_ar = None

  @abstractmethod
  def _AsyncProcessRequestImpl(self, sink_stack, msg, stream, headers):
    pass

  def AsyncProcessRequest(self, sink_stack, msg, stream, headers):
    if not self.__open_ar.ready():
      def _on_open_done(_):
        timeout_event = msg.properties.get(Deadline.EVENT_KEY, None)
        # Either there is no timeout, or the timeout hasn't expired yet.
        if not timeout_event or not timeout_event.Get():
          self._AsyncProcessRequestImpl(sink_stack, msg, stream, headers)
      self.__open_ar.rawlink(_on_open_done)
    else:
      self._AsyncProcessRequestImpl(sink_stack, msg, stream, headers)

  def __GetEndpoint(self, instance):
    if self._endpoint_name:
      aeps = instance.additional_endpoints
      ep = aeps.get(self._endpoint_name, None)
      if not ep:
        raise ValueError(
            "Endpoint name %s not found in endpoints", self._endpoint_name)
      return ep
    else:
      return instance.service_endpoint

  def _OnServersChanged(self, endpoint, channel_factory, added):
    """Overridable by child classes.  Invoked when servers in the server set are
    added or removed.

    Args:
      instance - The server set member being added or removed.
      added - True if the instance is being added, False if it's being removed.
    """
    pass

  def __AddServer(self, instance):
    """Adds a servers to the set of servers available to the load balancer.
    Note: The new sink is not opened at this time.

    Args:
      instance - A Member object to be added to the pool.
    """
    ep = self.__GetEndpoint(instance)
    if not ep in self._servers:
      new_props = self._properties.copy()
      new_props.update({ SinkProperties.Endpoint: ep })
      channel_factory = functools.partial(self._next_sink_provider.CreateSink, new_props)
      self._servers[ep] = channel_factory
      self._log.info("Instance %s joined (%d members)" % (
        ep, len(self._servers)))
      self._OnServersChanged(ep, channel_factory, True)

  def __RemoveServer(self, instance):
    """Removes a server from the load balancer.

    Args:
      instance - A Member object to be removed from the pool.
    """
    ep = self.__GetEndpoint(instance)
    channel_factory = self._servers.pop(ep, None)
    self._OnServersChanged(ep, channel_factory, False)

  def __OnServerSetJoin(self, instance):
    """Invoked when an instance joins the server set.

    Args:
      instance - Instance added to the cluster.
    """
    # callbacks from the ServerSet are delivered serially, so we can guarantee
    # that once this unblocks, we'll still get the notifications delivered in
    # the order that they arrived.  Ex: OnJoin(a) -> OnLeave(a)
    self.__init_done.wait()
    # OnJoin notifications are delivered at startup, however we already
    # pre-populate our copy of the ServerSet, so it's fine to ignore duplicates.
    if self.__GetEndpoint(instance) in self._servers:
      return

    self.__AddServer(instance)

  def __OnServerSetLeave(self, instance):
    """Invoked when an instance leaves the server set.

    Args:
      instance - Instance leaving the cluster.
    """
    self.__init_done.wait()
    self.__RemoveServer(instance)

    self._log.info("Instance %s left (%d members)" % (
        self.__GetEndpoint(instance), len(self._servers)))



"""Sinks are classes that control and modify the flow of messages through the
 RPC system.

 Sinks cooperatively chain together in a linked list.  Each sink in the chain
 calls the next sink until the chain terminates.  On the response side, sinks
 cooperatively propagate the response to the next via a sink stack.
"""

from abc import (
  ABCMeta,
  abstractmethod,
  abstractproperty
)
from collections import (deque, namedtuple)
from weakref import WeakValueDictionary
import time

try:
  from gevent.coros import RLock
except:
  from gevent.lock import RLock

from .asynchronous import AsyncResult
from .constants import (ChannelState, SinkProperties, SinkRole)
from .observable import Observable
from .message import (
  Deadline,
  MethodReturnMessage,
  TimeoutError
)
from .scales_socket import ScalesSocket
from .timer_queue import GLOBAL_TIMER_QUEUE
from .varz import (
  Rate,
  Source,
  VarzBase,
  VarzSocketWrapper
)

class MessageSink(ABCMeta('ABCMeta', (object,), {})):
  """A base class for all message sinks.

  MessageSinks form a cooperative linked list, which each sink calling the
  next sink in the chain once it's processing is complete.
  """
  __slots__ = '_next',

  def __init__(self):
    super(MessageSink, self).__init__()
    self._next = None

  @property
  def next_sink(self):
    """The next sink in the chain."""
    return self._next

  @next_sink.setter
  def next_sink(self, value):
    self._next = value


class ClientMessageSink(MessageSink):
  """ClientMessageSinks take a message, stream, and headers and perform
  processing on them.
  """
  __slots__ = '_on_faulted',
  Role = None
  Builder = None

  def __init__(self):
    self._on_faulted = Observable()
    super(ClientMessageSink, self).__init__()

  @property
  def state(self):
    return self.next_sink.state

  @property
  def is_open(self):
    """Returns True if the sink is Idle, Open, or Busy"""
    return self.state <= ChannelState.Busy

  @property
  def is_closed(self):
    """Returns True if the sink is Closed."""
    return self.state == ChannelState.Closed

  @property
  def is_ready(self):
    """Returns True if the channel is open, eg ready to process messages."""
    return self.state == ChannelState.Open

  @property
  def on_faulted(self):
    return self._on_faulted

  def Open(self):
    if self.next_sink:
      return self.next_sink.Open()
    else:
      return AsyncResult.Complete()

  def Close(self):
    if self.next_sink:
      self.next_sink.Close()

  @abstractmethod
  def AsyncProcessRequest(self, sink_stack, msg, stream, headers):
    """Process a request message, stream, and headers.

    Args:
      sink_stack - The SinkStack representing the processing state of the message.
                   Implementors should push their sink onto this stack before
                   forwarding the message in order to participate in processing
                   the response.
      msg - The message being processed.
      stream - A serialized version of the message.
      headers - Any additional headers to be sent.
    """
    raise NotImplementedError()

  @abstractmethod
  def AsyncProcessResponse(self, sink_stack, context, stream, msg):
    """Process a response stream.

    Args:
      sink_stack - The SinkStack representing the processing state of the message.
                   Implementors should call sink_stack.AsyncProcessMessage(...)
                   to forward the message to the next sink.
      context - The context that was pushed onto the stack in AsyncProcessRequest.
      stream - The stream representing the serialized response.
    """
    raise NotImplementedError()


class SinkStack(object):
  """A stack of sinks."""
  __slots__ = '_stack',

  def __init__(self):
    self._stack = deque()

  def Push(self, sink, context=None):
    """Push a sink, and optional context data, onto the stack.

    Args:
      sink - The sink to push onto the stack.
      context - Optional context data associated with the current processing
                state of the sink.
    """
    if sink is None:
      raise Exception("sink must not be None")

    self._stack.append((sink, context))

  def Pop(self):
    return self._stack.pop()

  def Any(self):
    return any(self._stack)


class ClientMessageSinkStack(SinkStack):
  """A SinkStack of ClientMessageSinks.

  The ClientMessageSinkStack forwards AsyncProcessResponse to the next sink
  on the stack.
  """

  def __init__(self):
    """
    Args:
      reply_sink - An optional ReplySink.
    """
    super(ClientMessageSinkStack, self).__init__()

  def AsyncProcessResponse(self, stream, msg):
    """Pop the next sink off the stack and call AsyncProcessResponse on it."""
    if self.Any():
      next_sink, next_ctx = self.Pop()
      next_sink.AsyncProcessResponse(self, next_ctx, stream, msg)

  def AsyncProcessResponseStream(self, stream):
    self.AsyncProcessResponse(stream, None)

  def AsyncProcessResponseMessage(self, msg):
    self.AsyncProcessResponse(None, msg)


class FailingMessageSink(ClientMessageSink):
  """A sink that always returns a failure message."""

  def __init__(self, ex):
    self._ex = ex
    super(FailingMessageSink, self).__init__()

  def AsyncProcessRequest(self, sink_stack, msg, stream, headers):
    msg = MethodReturnMessage(error=self._ex())
    sink_stack.AsyncProcessResponseMessage(msg)

  def AsyncProcessResponse(self, sink_stack, context, stream, msg):
    raise NotImplementedError("This should never be called")

  @property
  def state(self):
    return ChannelState.Open

  @property
  def endpoint(self):
    return None


class ClientTimeoutSink(ClientMessageSink):

  class Varz(VarzBase):
    _VARZ_BASE_NAME = 'scales.TimeoutSink'
    _VARZ = {
      'timeouts': Rate
    }

  def __init__(self, next_provider, sink_properties, global_properties):
    super(ClientTimeoutSink, self).__init__()
    self.next_sink = next_provider.CreateSink(global_properties)
    self._varz = self.Varz(
      Source(service=global_properties[SinkProperties.Label]))

  def _TimeoutHelper(self, evt, sink_stack):
    """Waits for ar to be signaled or [timeout] seconds to elapse.  If the
    timeout elapses, the event on the message will be signaled, and a timeout
    message posted to the sink_stack, aborting the message call.
    """
    if evt:
      evt.Set(True)
    self._varz.timeouts()
    error_msg = MethodReturnMessage(error=TimeoutError())
    sink_stack.AsyncProcessResponseMessage(error_msg)

  def AsyncProcessRequest(self, sink_stack, msg, stream, headers):
    """Initialize the timeout handler for this request.

    Args:
      ar - The AsyncResult for the pending response of this request.
      timeout - An optional timeout.  If None, no timeout handler is initialized.
      tag - The tag of the request.
    """
    deadline = msg.properties.get(Deadline.KEY)
    if deadline:
      now = time.time()
      if deadline < now:
        self._TimeoutHelper(None, sink_stack)
        return

      evt = Observable()
      msg.properties[Deadline.EVENT_KEY] = evt
      cancel_timeout = GLOBAL_TIMER_QUEUE.Schedule(
        deadline, lambda: self._TimeoutHelper(evt, sink_stack))
      sink_stack.Push(self, cancel_timeout)
    return self.next_sink.AsyncProcessRequest(sink_stack, msg, stream, headers)

  def AsyncProcessResponse(self, sink_stack, context, stream, msg):
    context()
    sink_stack.AsyncProcessResponse(stream, msg)


class SinkProviderBase(object):
  """Base class for sink providers."""
  __metaclass__ = ABCMeta
  __slots__ = 'next_provider', 'sink_properties'
  _defaults = {}
  SINK_CLASS = None
  PARAMS_CLASS = None
  Role = None

  def __init__(self, **kwargs):
    new_props = self._defaults.copy()
    new_props.update(kwargs)
    if self.PARAMS_CLASS:
      if not callable(self.PARAMS_CLASS):
        raise Exception('PARAMS_CLASS must be callable')
      self.sink_properties = self.PARAMS_CLASS(**new_props) # pylint: disable=E1102
    else:
      self.sink_properties = None
    self.next_provider = None

  @abstractmethod
  def CreateSink(self, properties):
    pass

  @abstractproperty
  def sink_class(self):
    pass

  def Clone(self, **kwargs):
    """Creates a copy of this sink_provider with new properties"""
    if hasattr(self.sink_properties, '_asdict'):
      new_props = self.sink_properties._asdict().copy()
    elif hasattr(self.sink_properties, '__dict__'):
      new_props = self.sink_properties.__dict__.copy()
    else:
      raise AttributeError('unable to copy sink_properties')
    new_props.update(kwargs)
    new_provider = type(self)(**new_props)
    new_provider.next_provider = self.next_provider
    return new_provider


def SinkProvider(sink_cls, role=None, **defaults):
  """Factory for creating simple sink providers.

  Args:
    sink_cls - The type of sink to provide.
  Returns:
    A SinkProvider that provides sinks of type 'sink_cls'.
  """
  field_names = ' '.join(defaults.keys())
  params_cls = namedtuple('Params', field_names)

  def CreateSink(self, properties):
    return self.SINK_CLASS(self.next_provider, self.sink_properties, properties)

  def sink_class(self):
    return self.SINK_CLASS

  provider = type(
    sink_cls.__name__ + 'Provider',
    (SinkProviderBase, ),
    {
      'SINK_CLASS': sink_cls,
      'PARAMS_CLASS': params_cls,
      'Role': role,
      'CreateSink': CreateSink,
      'sink_class': property(sink_class),
      '_defaults': defaults
    }
  )
  return provider


TimeoutSinkProvider = SinkProvider(ClientTimeoutSink)


def SocketTransportSinkProvider(sink_cls):
  class _SocketTransportSinkProvider(SinkProviderBase):
    SINK_CLS = sink_cls
    Role = SinkRole.Transport

    def CreateSink(self, properties):
      server = properties[SinkProperties.Endpoint]
      service = properties[SinkProperties.Label]
      sock = ScalesSocket(server.host, server.port)
      healthy_sock = VarzSocketWrapper(sock, service)
      sink = self.SINK_CLS(healthy_sock, service)
      return sink

    @property
    def sink_class(self):
      return self.SINK_CLS

  return _SocketTransportSinkProvider


class RefCountedSink(ClientMessageSink):
  def __init__(self, next_sink):
    super(RefCountedSink, self).__init__()
    self._ref_count = 0
    self._open_lock = RLock()
    self._open_ar = None
    self.next_sink = next_sink

  def AsyncProcessRequest(self, sink_stack, msg, stream, headers):
    self.next_sink.AsyncProcessRequest(sink_stack, msg, stream, headers)

  def AsyncProcessResponse(self, sink_stack, context, stream, msg):
    raise NotImplementedError("Not called")

  @property
  def on_faulted(self):
    return self.next_sink.on_faulted

  def Open(self):
    with self._open_lock:
      self._ref_count += 1
      if self._ref_count == 1:
        self._open_ar = self.next_sink.Open()
    return self._open_ar

  def Close(self):
    with self._open_lock:
      if self._ref_count == 0:
        return
      self._ref_count -= 1
      if self._ref_count == 0:
        self._open_ar = None
        self.next_sink.Close()


class SharedSinkProvider(SinkProviderBase):
  def __init__(self, key_selector):
    self._key_selector = key_selector
    self._cache = WeakValueDictionary()
    super(SharedSinkProvider, self).__init__()

  def CreateSink(self, properties):
    key = self._key_selector(properties)
    if key:
      sink = self._cache.get(key)
      if not sink:
        new_sink = self.next_provider.CreateSink(properties)
        sink = RefCountedSink(new_sink)
        self._cache[key] = sink
      return sink
    else:
      return self.next_provider.CreateSink(properties)

  @property
  def sink_class(self):
    return self.next_provider.sink_class



import time

import gevent
import redis

from ..asynchronous import (
  AsyncResult,
  NoopTimeout
)
from ..sink import (
  ClientMessageSink,
  SinkProvider
)
from ..constants import (
  ChannelState,
  SinkProperties,
)
from ..message import (
  ChannelConcurrencyError,
  Deadline,
  MethodReturnMessage,
  TimeoutError
)
from ..varz import (
  AverageTimer,
  Rate,
  Source,
  VarzBase,
)

class RedisTransportSink(ClientMessageSink):
  class Varz(VarzBase):
    _VARZ_BASE_NAME = 'scales.redis.RedisTransportSink'
    _VARZ = {
      'messages_sent': Rate,
      'messages_recv': Rate,
      'transport_latency': AverageTimer
    }

  def __init__(self, next_provider, sink_properties, global_properties):
    super(RedisTransportSink, self).__init__()
    self._endpoint = global_properties[SinkProperties.Endpoint]
    self._state = ChannelState.Idle
    name = global_properties[SinkProperties.Label]
    socket_source = '%s:%d' % (self._endpoint.host, self._endpoint.port)
    self._varz = self.Varz(Source(service=name, endpoint=socket_source))
    self._open_result = None
    self._client = None
    self._processing = None

  def Open(self):
    if not self._open_result:
      self._open_result = AsyncResult()
      self._open_result.SafeLink(self._OpenImpl)
    return self._open_result

  def _OpenImpl(self):
    self._client = redis.StrictRedis(
        host=self._endpoint.host,
        port=self._endpoint.port)
    try:
      self._client.ping()
      self._state = ChannelState.Open
    except Exception as ex:
      self._Fault(ex)
      raise

  def Close(self):
    self._state = ChannelState.Closed
    if self._client:
      client, self._client = self._client, None
      client.connection_pool.disconnect()
    self._open_result = None
    if self._processing:
      p, self._processing = self._processing, None
      p.kill(block=False)

  def _Fault(self, reason):
    if self.state == ChannelState.Closed:
      return

    self.Close()
    if not isinstance(reason, Exception):
      reason = Exception(str(reason))
    self._on_faulted.Set(reason)

  @property
  def state(self):
    return self._state

  def _AsyncProcessTransaction(self, sink_stack, msg, deadline):
    method = getattr(self._client, msg.method)
    with self._varz.transport_latency.Measure():
      gtimeout = None
      try:
        if deadline:
          timeout = deadline - time.time()
          if timeout < 0:
            raise gevent.Timeout()
          gtimeout = gevent.Timeout.start_new(timeout)
        else:
          gtimeout = NoopTimeout()

        self._varz.messages_sent()
        ret = method(*msg.args, **msg.kwargs)
        msg = MethodReturnMessage(ret)
        gtimeout.cancel()
        self._processing = None
        gevent.spawn(sink_stack.AsyncProcessResponseMessage, msg)
      except gevent.Timeout: # pylint: disable=E0712
        err = TimeoutError()
        self._client.connection_pool.disconnect()
        self._processing = None
        gevent.spawn(sink_stack.AsyncProcessResponseMessage, MethodReturnMessage(error=err))
      except Exception as err:
        self._processing = None
        if gtimeout:
          gtimeout.cancel()
        self._Fault(err)
        gevent.spawn(sink_stack.AsyncProcessResponseMessage, MethodReturnMessage(error=err))
      finally:
        self._varz.messages_recv()

  def AsyncProcessRequest(self, sink_stack, msg, stream, headers):
    if self._processing:
      sink_stack.AsyncProcessResponseMessage(MethodReturnMessage(
        error=ChannelConcurrencyError(
          'Concurrency violation in AsyncProcessRequest')))
      return
    deadline = msg.properties.get(Deadline.KEY)
    self._processing = gevent.spawn(self._AsyncProcessTransaction, sink_stack, msg, deadline)

  def AsyncProcessResponse(self, sink_stack, context, stream, msg):
    pass

RedisTransportSink.Builder = SinkProvider(RedisTransportSink)



from struct import (pack, unpack)
import time

import gevent
from thrift.protocol.TBinaryProtocol import TBinaryProtocolAcceleratedFactory

from ..asynchronous import (
  AsyncResult,
  NoopTimeout
)
from ..compat import BytesIO
from ..constants import (
  ChannelState,
  SinkProperties,
  SinkRole
)
from ..message import (
  ChannelConcurrencyError,
  ClientError,
  Deadline,
  MethodCallMessage,
  MethodReturnMessage,
  TimeoutError
)
from ..sink import (
  SinkProvider,
  SocketTransportSinkProvider,
  ClientMessageSink,
)
from ..varz import (
  AggregateTimer,
  AverageTimer,
  Rate,
  Source,
  VarzBase,
)
from .serializer import MessageSerializer


class SocketTransportSink(ClientMessageSink):
  """A sink to transport thrift method calls over a socket.

  This sink does not support processing multiple messages in parallel and will
  raise an exception if it detects it is about to.
  """
  class Varz(VarzBase):
    """
    messages_sent - The number of messages sent over this sink.
    messages_recv - The number of messages received over this sink.
    send_time - The aggregate amount of time spent sending data.
    recv_time - The aggregate amount of time spend receiving data.
    send_latency - The average amount of time taken to send a message.
    recv_latency - The average amount of time taken to receive a message
                   (once a response has reached the client).
    transport_latency - The average amount of time taken to perform a full
                        method call transaction (send data, wait for response,
                        read response).
    """

    _VARZ_BASE_NAME = 'scales.thrift.SocketTransportSink'
    _VARZ = {
      'messages_sent': Rate,
      'messages_recv': Rate,
      'send_time': AggregateTimer,
      'recv_time': AggregateTimer,
      'send_latency': AverageTimer,
      'recv_latency': AverageTimer,
      'transport_latency': AverageTimer
    }

  def __init__(self, socket, source):
    super(SocketTransportSink, self).__init__()
    self._socket = socket
    self._state = ChannelState.Idle
    socket_source = '%s:%d' % (self._socket.host, self._socket.port)
    self._varz = self.Varz(Source(service=source, endpoint=socket_source))
    self._processing = None
    self._open_result = None

  def Open(self):
    if not self._open_result:
      self._open_result = AsyncResult()
      self._open_result.SafeLink(self._OpenImpl)
    return self._open_result

  def _OpenImpl(self):
    try:
      self._socket.open()
      self._state = ChannelState.Open
    except Exception:
      self._Fault('Open failed')
      raise

  def Close(self):
    self._state = ChannelState.Closed
    self._socket.close()
    self._open_result = None
    if self._processing:
      p, self._processing = self._processing, None
      p.kill(block=False)

  @property
  def state(self):
    if self._socket.isOpen():
      return ChannelState.Open
    else:
      return self._state

  def _Fault(self, reason):
    """Shutdown the sink and signal.

    Args:
      reason - The reason the shutdown occurred.  May be an exception or string.
    """
    if self.state == ChannelState.Closed:
      return

    self.Close()
    if not isinstance(reason, Exception):
      reason = Exception(str(reason))
    self._on_faulted.Set(reason)

  def _AsyncProcessTransaction(self, data, sink_stack, deadline):
    """Process a method call/response transaction.
    When complete the response is dispatched to the sink_stack.

    Args:
      data - The data to send.
      sink_stack - The sink_stack for this method call.
      deadline - The absolute deadline the transaction must complete by.
    """
    with self._varz.transport_latency.Measure():
      gtimeout = None
      try:
        if deadline:
          timeout = deadline - time.time()
          if timeout < 0:
            raise gevent.Timeout()
          gtimeout = gevent.Timeout.start_new(timeout)
        else:
          gtimeout = NoopTimeout()

        with self._varz.send_time.Measure():
          with self._varz.send_latency.Measure():
            self._socket.write(data)
        self._varz.messages_sent()

        sz, = unpack('!i', self._socket.readAll(4))
        with self._varz.recv_time.Measure():
          with self._varz.recv_latency.Measure():
            buf = BytesIO(self._socket.readAll(sz))
        self._varz.messages_recv()

        gtimeout.cancel()
        self._processing = None
        gevent.spawn(self._ProcessReply, buf, sink_stack)
      except gevent.Timeout: # pylint: disable=E0712
        err = TimeoutError()
        self._socket.close()
        self._socket.open()
        self._processing = None
        sink_stack.AsyncProcessResponseMessage(MethodReturnMessage(error=err))
      except Exception as ex:
        if gtimeout:
          gtimeout.cancel()
        self._Fault(ex)
        self._processing = None
        sink_stack.AsyncProcessResponseMessage(MethodReturnMessage(error=ex))

  @staticmethod
  def _ProcessReply(buf, sink_stack):
    try:
      buf.seek(0)
      sink_stack.AsyncProcessResponseStream(buf)
    except Exception as ex:
      sink_stack.AsyncProcessResponseMessage(MethodReturnMessage(error=ex))

  def AsyncProcessRequest(self, sink_stack, msg, stream, headers):
    if self._processing is not None:
      sink_stack.AsyncProcessResponseMessage(MethodReturnMessage(
        error=ChannelConcurrencyError(
          'Concurrency violation in AsyncProcessRequest')))
      return

    payload = stream.getvalue()
    sz = pack('!i', len(payload))
    deadline = msg.properties.get(Deadline.KEY)
    self._processing = gevent.spawn(self._AsyncProcessTransaction, sz + payload, sink_stack, deadline)

  def AsyncProcessResponse(self, sink_stack, context, stream, msg):
    pass


SocketTransportSink.Builder = SocketTransportSinkProvider(SocketTransportSink)


class ThriftSerializerSink(ClientMessageSink):
  """A sink for serializing thrift method calls."""

  def __init__(self, next_provider, sink_properties, global_properties):
    super(ThriftSerializerSink, self).__init__()
    self._serializer = MessageSerializer(
        global_properties[SinkProperties.ServiceInterface],
        sink_properties.protocol_factory)
    self.next_sink = next_provider.CreateSink(global_properties)

  def AsyncProcessRequest(self, sink_stack, msg, stream, headers):
    buf = BytesIO()
    headers = {}

    if not isinstance(msg, MethodCallMessage):
      sink_stack.AsyncProcessResponseMessage(MethodReturnMessage(
          error=ClientError('Invalid message type.')))
      return
    try:
      self._serializer.SerializeThriftCall(msg, buf)
    except Exception as ex:
      sink_stack.AsyncProcessResponseMessage(MethodReturnMessage(error=ex))
      return

    sink_stack.Push(self)
    self.next_sink.AsyncProcessRequest(sink_stack, msg, buf, headers)

  def AsyncProcessResponse(self, sink_stack, context, stream, msg):
    if msg:
      # No need to deserialize, it already is.
      sink_stack.AsyncProcessResponseMessage(msg)
    else:
      try:
        msg = self._serializer.DeserializeThriftCall(stream)
      except Exception as ex:
        msg = MethodReturnMessage(error=ex)
      sink_stack.AsyncProcessResponseMessage(msg)


ThriftSerializerSink.Builder = SinkProvider(
  ThriftSerializerSink,
  SinkRole.Formatter,
  protocol_factory=TBinaryProtocolAcceleratedFactory())



from .base import PoolSink
from ..asynchronous import AsyncResult
from ..constants import (ChannelState, SinkRole)
from ..sink import SinkProvider, SinkProperties


class SingletonPoolSink(PoolSink):
  """A SingletonPool maintains at most one underlying sink, and allows concurrent
  requests through it.  If the underlying sink fails, it's closed and reopened.

  This sink is intended to be used for transports that can handle concurrent
  requests, such as multiplexing transports.
  """

  def __init__(self, sink_provider, sink_properties, global_properties):
    super(SingletonPoolSink, self).__init__(sink_provider, global_properties)
    endpoint = global_properties[SinkProperties.Endpoint]
    endpoint_source = '%s:%d' % (endpoint.host, endpoint.port)
    self.endpoint = endpoint_source
    self._ref_count = 0

  def Open(self):
    """Open the underlying sink and increase the ref count."""
    self._ref_count += 1
    if self._ref_count > 1:
      return AsyncResult.Complete()

    def TryGet():
      self._Get()
      return True
    # We don't want to link _Get directly as it'll hold a reference
    # to the sink returned forever.
    return AsyncResult.Run(TryGet)

  def Close(self):
    """Decrease the reference count and, if zero, close the underlying transport."""
    self. _ref_count -= 1
    if self.next_sink and self._ref_count <= 0:
      sink, self.next_sink = self.next_sink, None
      sink.on_faulted.Unsubscribe(self.__PropagateShutdown)
      sink.Close()

  @property
  def state(self):
    if self.next_sink:
      return self.next_sink.state
    else:
      return ChannelState.Idle

  def __PropagateShutdown(self, value):
    self.on_faulted.Set(value)

  def _Get(self):
    if not self.next_sink:
      self.next_sink = self._sink_provider.CreateSink(self._properties)
      self.next_sink.on_faulted.Subscribe(self.__PropagateShutdown)
      self.next_sink.Open().wait()
      return self.next_sink
    elif self.next_sink.state == ChannelState.Idle:
      self.next_sink.Open().wait()
      return self.next_sink
    elif self.next_sink.is_closed:
      self.next_sink.on_faulted.Unsubscribe(self.__PropagateShutdown)
      self.next_sink = None
      return self._Get()
    else:
      return self.next_sink

  def _Release(self, sink):
    pass


SingletonPoolSink.Builder = SinkProvider(SingletonPoolSink, SinkRole.Pool)



from thrift.protocol.TJSONProtocol import TJSONProtocolFactory

from .sink import (
  ThriftHttpTransportSink
)
from ..core import Scales
from ..loadbalancer.aperture import ApertureBalancerSink
from ..pool import SingletonPoolSink
from ..thrift.sink import ThriftSerializerSink

class ThriftHttp(object):
  @staticmethod
  def NewBuilder(Iface, url_suffix):
    return Scales.NewBuilder(Iface) \
      .WithSink(ThriftSerializerSink.Builder(protocol_factory=TJSONProtocolFactory())) \
      .WithSink(ApertureBalancerSink.Builder()) \
      .WithSink(SingletonPoolSink.Builder()) \
      .WithSink(ThriftHttpTransportSink.Builder(url=url_suffix))

  @staticmethod
  def NewClient(Iface, uri, timeout=60, url_suffix=''):
    return ThriftHttp.NewBuilder(Iface, url_suffix) \
      .SetUri(uri) \
      .SetTimeout(timeout) \
      .Build()



from abc import (ABCMeta, abstractmethod)
from six import string_types

class ServerSetProvider(ABCMeta('ABCMeta', (object,), {})):
  """Base class for providing a set of servers, as well as optionally
  notifying the pool of servers joining and leaving the set."""

  @abstractmethod
  def Initialize(self, on_join, on_leave):
    """Initialize the provider.
    This method is called before any calls to GetServers().

     Args:
      on_join - A function to be called when a server joins the set.
      on_leave - A function to be called when a server leaves the set.
    """
    raise NotImplementedError()

  @abstractmethod
  def Close(self):
    """Close the provider and any resources associated.
    """
    raise NotImplementedError()

  @abstractmethod
  def GetServers(self):
    """Get all the current servers in the set.

    Returns:
      An iterable of servers.
    """
    raise NotImplementedError()

  @property
  def endpoint_name(self):
    return None


class StaticServerSetProvider(ServerSetProvider):
  """A ServerSetProvider that returns a static set of servers."""

  def __init__(self, servers):
    """Initializes the set with a static list of servers.

    Args:
      servers - An iterable of servers.
    """
    self._servers = servers

  def Initialize(self, on_join, on_leave):
    pass

  def Close(self):
    pass

  def GetServers(self):
    return self._servers


class ZooKeeperServerSetProvider(ServerSetProvider):
  """A ServerSetProvider that tracks servers in zookeeper."""
  from ..loadbalancer.zookeeper import ServerSet
  from kazoo.client import KazooClient
  from kazoo.handlers.gevent import SequentialGeventHandler

  def __init__(self,
      zk_servers_or_client,
      zk_path,
      zk_timeout=30.0,
      member_prefix='member_',
      member_factory=None,
      endpoint_name=None):
    """
    Args:
        zk_servers_or_client - Either a comma separated list of host:port ZK
                               servers, or an instance of a KazooClient.
        zk_path - The ZK path to discover services from.
        zk_timeout - The timeout to set on the ZK client.  Ignored if
                     zk_servers_or_client is a KazooClient.
        member_prefix - The prefix to match for children in the watched znode.
        member_factory - A callable to produce a Member object from a znode.
        endpoint_name - The name of the endpoint on the Member.  If None, the
                        endpoint is the serviceEndpoint in the znode data,
                        otherwise it is taken from the additionalEndpoints data
                        in the znode.
    """
    self._zk_client = None
    if isinstance(zk_servers_or_client, string_types):
      self._zk_client = self._GetZooKeeperClient(zk_servers_or_client, zk_timeout)
      self._owns_zk_client = True
    else:
      self._zk_client = zk_servers_or_client
      self._owns_zk_client = False
    self._zk_path = zk_path
    self._server_set = None
    self._member_prefix = member_prefix
    self._member_factory = member_factory
    self._endpoint_name = endpoint_name

  def _GetZooKeeperClient(self, zk_servers, zk_timeout):
    if self._zk_client:
      return self._zk_client
    else:
      return self.KazooClient(
        hosts=zk_servers,
        timeout=zk_timeout,
        handler=self.SequentialGeventHandler(),
        randomize_hosts=True)

  def _MemberFilter(self, node):
    return node.startswith(self._member_prefix)

  def Initialize(self, on_join, on_leave):
    self._zk_client.start()
    self._server_set = self.ServerSet(
      self._zk_client, self._zk_path, on_join, on_leave,
      self._MemberFilter, self._member_factory)

  def Close(self):
    if self._server_set:
      self._server_set.stop()
    if self._owns_zk_client and self._zk_client:
      self._zk_client.stop()

  def GetServers(self):
    if not self._server_set:
      raise Exception('Initialize() must be called first.')
    return self._server_set.get_members()

  @property
  def endpoint_name(self):
    return self._endpoint_name



"""Core classes for dispatching messages from a Scales proxy to a message sink stack."""

import time

import gevent

from .asynchronous import AsyncResult
from .constants import MessageProperties, SinkProperties
from .message import (
  Deadline,
  MethodCallMessage,
  MethodReturnMessage,
  TimeoutError
)
from .sink import (
  ClientMessageSink,
  ClientMessageSinkStack
)
from .varz import (
  Rate,
  Source,
  AverageTimer,
  VarzBase
)


class InternalError(Exception):
  pass


class ScalesError(Exception):
  def __init__(self, ex, msg):
    self.inner_exception = ex
    super(ScalesError, self).__init__(msg)


class ServiceClosedError(Exception):
  pass


class _AsyncResponseSink(ClientMessageSink):
  @staticmethod
  def _WrapException(msg):
    """Creates an exception object that contains the inner exception
    from a SystemErrorMessage.  This allows the actual failure stack
    to propagate to the waiting greenlet.

    Args:
      msg - The MethodReturnMessage that has an active error.

    Returns:
      An exception object wrapping the error in the MethodCallMessage.
    """
    # Don't wrap timeouts.
    if isinstance(msg.error, TimeoutError):
      return msg.error

    stack = getattr(msg, 'stack', None)
    if stack:
      ex_msg = """An error occurred while processing the request
[Inner Exception: --------------------]
%s[End of Inner Exception---------------]
""" % ''.join(stack)
      return ScalesError(msg.error, ex_msg)
    else:
      return msg.error

  def AsyncProcessRequest(self, sink_stack, msg, stream, headers):
    raise NotImplementedError("This should never be called")

  def AsyncProcessResponse(self, sink_stack, context, stream, msg):
    """Propagate the results from a message onto an AsyncResult.

    Args:
      msg - The reply message (a MethodReturnMessage).
    """
    source, start_time, ar, msg_props = context
    endpoint = msg_props.get(MessageProperties.Endpoint, None)
    if endpoint and not isinstance(endpoint, str):
      endpoint = str(endpoint)
    if source:
      host_source = Source(method=source.method,
        service=source.service,
        endpoint=endpoint)
    else:
      host_source = None

    end_time = time.time()
    if host_source:
      MessageDispatcher.Varz.request_latency(host_source, end_time - start_time) # pylint: disable=no-member
    if isinstance(msg, MethodReturnMessage):
      if msg.error:
        if host_source:
          MessageDispatcher.Varz.exception_messages(host_source) # pylint: disable=no-member
        ar.set_exception(self._WrapException(msg))
      else:
        if host_source:
          MessageDispatcher.Varz.success_messages(host_source) # pylint: disable=no-member
        ar.set(msg.return_value)
    else:
      ar.set_exception(InternalError('Unknown response message of type %s'
                                     % msg.__class__))


class MessageDispatcher(ClientMessageSink):
  """Handles dispatching incoming and outgoing messages to a client sink stack."""

  class Varz(VarzBase):
    """
    dispatch_messages - The number of messages dispatched.
    success_messages - The number of successful responses processed.
    exception_messages - The number of exception responses processed.
    request_latency - The average time taken to receive a response to a request.
    """
    _VARZ_BASE_NAME = 'scales.MessageDispatcher'
    _VARZ = {
      'dispatch_messages': Rate,
      'success_messages': Rate,
      'exception_messages': Rate,
      'request_latency': AverageTimer
    }

  def __init__(
        self,
        service,
        sink_provider,
        default_timeout,
        properties):
    """
    Args:
      service - The service interface class this dispatcher is serving.
      sink_provider - An instance of a SinkProvider.
      properties - The properties associated with this service and dispatcher.
    """
    super(MessageDispatcher, self).__init__()
    self.next_sink = sink_provider.CreateSink(properties)
    self._dispatch_timeout = default_timeout
    self._service = service
    self._name = properties[SinkProperties.Label]
    self._open_ar = AsyncResult()

  def Open(self):
    self._open_ar = self.next_sink.Open()
    return self._open_ar

  def Close(self):
    self.next_sink.Close()
    self._open_ar = None

  def DispatchMethodCall(self, method, args, kwargs, timeout=None):
    """Creates and posts a Tdispatch message to a client sink stack.

    Args:
      service - The service interface originating this call.
      method  - The method being called.
      args    - The parameters passed to the method.
      kwargs  - The keyword parameters passed to the method.
      timeout - An optional timeout.  If not set, the global dispatch timeout
                will be applied.

    Returns:
      An AsyncResult representing the status of the method call.  This will be
      signaled when either the call completes (successfully or from failure),
      or after [timeout] seconds elapse.
    """
    if not self._open_ar:
      raise Exception('Dispatcher not open.')

    timeout = timeout or self._dispatch_timeout
    start_time = time.time()
    if self._open_ar.ready():
      return self._DispatchMethod(method, args, kwargs, timeout, start_time)
    else:
      # _DispatchMethod returns an AsyncResult, so we end up with an
      # AsyncResult<AsyncResult<TRet>>, Unwrap() removes one layer, yielding
      # an AsyncResult<TRet>
      return self._open_ar.ContinueWith(
          lambda ar: self._DispatchMethod(method, args, kwargs, timeout, start_time)
      ).Unwrap()

  @staticmethod
  def StaticDispatchMessage(sink, source, start_time, deadline, disp_msg):
    # Init the properties dictionary w/ an empty endpoint
    disp_msg.properties[MessageProperties.Endpoint] = None
    if deadline:
      disp_msg.properties[Deadline.KEY] = deadline

    ar = AsyncResult()
    sink_stack = ClientMessageSinkStack()
    repsonse_sink = _AsyncResponseSink()
    sink_stack.Push(repsonse_sink, (source, start_time, ar, disp_msg.properties))
    gevent.spawn(sink.AsyncProcessRequest, sink_stack, disp_msg, None, {})
    return ar

  def _DispatchMethod(self, method, args, kwargs, timeout, start_time):
    open_time = time.time()
    open_latency = open_time - start_time

    if timeout:
      # Calculate the deadline for this method call.
      # Reduce it by the time it took for the open() to complete.
      deadline = start_time + timeout - open_latency
    else:
      deadline = None

    disp_msg = MethodCallMessage(self._service, method, args, kwargs)
    source = Source(method=method, service=self._name)
    self.Varz.dispatch_messages(source) # pylint: disable=no-member
    return self.StaticDispatchMessage(self.next_sink, source, start_time, deadline, disp_msg)

  def AsyncProcessRequest(self, sink_stack, msg, stream, headers):
    raise NotImplementedError("This should never be called.")

  def AsyncProcessResponse(self, sink_stack, context, stream, msg):
    raise NotImplementedError("This should never be called.")



import inspect
import sys

from thrift.protocol.TBinaryProtocol import TBinaryProtocolAcceleratedFactory
from thrift.transport.TTransport import TMemoryBuffer
from thrift.Thrift import (
  TApplicationException,
  TMessageType
)

from ..message import MethodReturnMessage


class MessageSerializer(object):
  """A serializer that can serialize and deserialize thrift method calls.

  This relies on the generated thrift args and return value classes created
  by the thrift compiler to do the serialization/deserialization.
  """
  def __init__(
      self,
      service_cls,
      protocol_factory=TBinaryProtocolAcceleratedFactory()):
    """
    Args:
      service_cls - The thrift generated interface class.
      protocol_factory - A class implementing getProtocol(...).  By default,
       TBinaryProtocolAcceleratedFactory is used.
    """
    self._protocol_factory = protocol_factory
    self._seq_id = 0
    self._service_modules = [sys.modules[c.__module__]
                             for c in inspect.getmro(service_cls)
                             if not c is object]
    if len(self._service_modules) == 1:
      self._FindClass = self._FindClassNoInheritance
    else:
      self._attr_cache = {}
      self._FindClass = self._FindClassInheritance

  def _FindClassInheritance(self, name):
    cls = self._attr_cache.get(name)
    if cls:
      return cls

    for m in self._service_modules:
      cls = getattr(m, name, None)
      if cls:
        self._attr_cache[name] = cls
        return cls
    return None

  def _FindClassNoInheritance(self, name):
    return getattr(self._service_modules[0], name, None)

  def SerializeThriftCall(self, msg, buf):
    """Serialize a MethodCallMessage to a stream

    Args:
      msg - The MethodCallMessage to serialize.
      buf - The buffer to serialize into.
    """
    thrift_buffer = TMemoryBuffer()
    thrift_buffer._buffer = buf
    protocol = self._protocol_factory.getProtocol(thrift_buffer)
    method, args, kwargs = msg.method, msg.args, msg.kwargs
    is_one_way = self._FindClass('%s_result' % method) is None
    args_cls = self._FindClass('%s_args' % method)
    if not args_cls:
      raise AttributeError('Unable to find args class for method %s' % method)

    protocol.writeMessageBegin(
        msg.method,
        TMessageType.ONEWAY if is_one_way else TMessageType.CALL,
        self._seq_id)
    thrift_args = args_cls(*args, **kwargs)
    thrift_args.write(protocol)
    protocol.writeMessageEnd()

  def DeserializeThriftCall(self, buf):
    """Deserialize a stream and context to a MethodReturnMessage.

    Args:
      buf - The buffer.
      ctx - The context from serialization.

    Returns:
      A MethodCallMessage.
    """

    thrift_buffer = TMemoryBuffer()
    thrift_buffer._buffer = buf
    protocol = self._protocol_factory.getProtocol(thrift_buffer)

    (fn_name, msg_type, seq_id) = protocol.readMessageBegin()
    if msg_type == TMessageType.EXCEPTION:
      x = TApplicationException()
      x.read(protocol)
      protocol.readMessageEnd()
      return MethodReturnMessage(error=x)

    result_cls = self._FindClass('%s_result' % fn_name)
    if result_cls:
      result = result_cls()
      result.read(protocol)
    else:
      result = None
    protocol.readMessageEnd()

    if not result:
      return MethodReturnMessage()
    if getattr(result, 'success', None) is not None:
      return MethodReturnMessage(return_value=result.success)

    result_spec = getattr(result_cls, 'thrift_spec', None)
    if result_spec:
      exceptions = result_spec[1:]
      for e in exceptions:
        attr_val = getattr(result, e[2], None)
        if attr_val is not None:
          return MethodReturnMessage(error=attr_val)

    return MethodReturnMessage(TApplicationException(
      TApplicationException.MISSING_RESULT, "%s failed: unknown result" % fn_name))




import logging
import random
import time
from struct import (pack, unpack)

import gevent

from ..asynchronous import AsyncResult
from ..compat import BytesIO
from ..constants import SinkProperties
from ..message import (
  Deadline,
  MethodDiscardMessage,
  MethodReturnMessage
)
from ..sink import (
  ClientMessageSink,
  SinkProvider,
  SocketTransportSinkProvider
)
from ..mux.sink import MuxSocketTransportSink
from ..varz import (
  AverageRate,
  Counter,
  Source,
  VarzBase
)
from .serializer import (
  MessageSerializer,
)
from .protocol import (
  MessageType,
)

ROOT_LOG = logging.getLogger('scales.thriftmux')


class SocketTransportSink(MuxSocketTransportSink):
  def __init__(self, socket, service):
    self._ping_timeout = 5
    self._ping_msg = self._BuildHeader(1, MessageType.Tping, 0)
    self._last_ping_start = 0
    super(SocketTransportSink, self).__init__(socket, service)

  def _Init(self):
    self._ping_ar = None
    super(SocketTransportSink, self)._Init()

  @staticmethod
  def _EncodeTag(tag):
    return [tag >> 16 & 0xff, tag >> 8 & 0xff, tag & 0xff] # Tag

  def _BuildHeader(self, tag, msg_type, data_len):
    total_len = 1 + 3 + data_len
    return pack('!ibBBB',
      total_len,
      msg_type,
      *self._EncodeTag(tag))

  def _PingLoop(self):
    """Periodically pings the remote server."""
    while self.isActive:
      gevent.sleep(random.randint(30, 40))
      if self.isActive:
        self._SendPingMessage()
      else:
        break

  def _SendPingMessage(self):
    """Constructs and sends a Tping message."""
    self._log.debug('Sending ping message.')
    self._ping_ar = AsyncResult()
    self._last_ping_start = time.time()
    self._send_queue.put((self._ping_msg, self._EMPTY_DCT))
    gevent.spawn(self._PingTimeoutHelper)
    return self._ping_ar

  def _OnPingResponse(self, msg_type, stream):
    """Handles the response to a ping.  On failure, shuts down the dispatcher.
    """
    ar, self._ping_ar = self._ping_ar, None
    if msg_type == MessageType.Rping:
      ar.set()
      ping_duration = time.time() - self._last_ping_start
      self._log.debug('Got ping response in %d ms' % int(ping_duration * 1000))
    else:
      self._log.error('Unexpected response for tag 1 (msg_type was %d)' % msg_type)
      ar.set_exception(Exception("Invalid ping response"))

  def _PingTimeoutHelper(self):
    ar = self._ping_ar
    ar.wait(self._ping_timeout)
    if not ar.successful():
      ar.set_exception(Exception('Ping timed out'))
      self._Shutdown('Ping Timeout')

  def _CheckInitialConnection(self):
    ar = self._SendPingMessage()
    ar.get()
    self._log.debug('Ping successful')
    self._greenlets.append(self._SpawnNamedGreenlet('Ping Loop', self._PingLoop))

  @staticmethod
  def _CreateDiscardMessage(tag):
    """Create a Tdiscarded message for 'tag'

    Args:
      tag - The message tag to discard.
    Returns
      A (message, buffer, headers) tuple suitable for passing to AsyncProcessRequest.
    """
    discard_message = MethodDiscardMessage(tag, 'Client timeout')
    discard_message.which = tag
    buf = BytesIO()
    headers = {}
    MessageSerializer(None).Marshal(discard_message, buf, headers)
    return discard_message, buf, headers

  def _OnTimeout(self, tag):
    if tag:
      msg, buf, headers = self._CreateDiscardMessage(tag)
      self.AsyncProcessRequest(None, msg, buf, headers)

  def _ProcessReply(self, stream):
    try:
      msg_type, tag = ThriftMuxMessageSerializerSink.ReadHeader(stream)
      if tag == 1 and msg_type == MessageType.Rping: #Ping
        self._OnPingResponse(msg_type, stream)
      elif tag != 0:
        self._ProcessTaggedReply(tag, stream)
      else:
        self._log.error('Unexpected message, msg_type = %d, tag = %d' % (msg_type, tag))
    except Exception:
      self._log.exception('Exception processing reply message.')

  def _Shutdown(self, reason, fault=True):
    super(SocketTransportSink, self)._Shutdown(reason, fault)
    if self._ping_ar:
      self._ping_ar.set_exception(reason)


SocketTransportSink.Builder = SocketTransportSinkProvider(SocketTransportSink)


class ThriftMuxMessageSerializerSink(ClientMessageSink):
  """A serializer sink that serializes thrift messages to the finagle mux
  wire format"""

  class Varz(VarzBase):
    _VARZ_BASE_NAME = 'scales.thriftmux.ThriftMuxMessageSerializerSink'
    _VARZ = {
      'deserialization_failures': Counter,
      'serialization_failures': Counter,
      'message_bytes_sent': AverageRate,
      'message_bytes_recv': AverageRate
    }

  def __init__(self, next_provider, sink_properties, global_properties):
    super(ThriftMuxMessageSerializerSink, self).__init__()
    self.next_sink = next_provider.CreateSink(global_properties)
    self._serializer = MessageSerializer(global_properties[SinkProperties.ServiceInterface])
    self._varz = self.Varz(Source(
      service=global_properties[SinkProperties.Label]))

  @staticmethod
  def ReadHeader(stream):
    """Read a mux header off a message.

    Args:
      msg - a byte buffer of raw data.

    Returns:
      A tuple of (message_type, tag)
    """
    header, = unpack('!i', stream.read(4))
    msg_type = (256 - (header >> 24 & 0xff)) * -1
    tag = ((header << 8) & 0xFFFFFFFF) >> 8
    return msg_type, tag

  def AsyncProcessRequest(self, sink_stack, msg, stream, headers):
    buf = BytesIO()
    headers = {}

    deadline = msg.properties.get(Deadline.KEY)
    if deadline:
      headers['com.twitter.finagle.Deadline'] = Deadline(deadline)

    try:
      self._serializer.Marshal(msg, buf, headers)
    except Exception as ex:
      self._varz.serialization_failures()
      msg = MethodReturnMessage(error=ex)
      sink_stack.AsyncProcessResponseMessage(msg)
      return

    self._varz.message_bytes_sent(buf.tell())
    sink_stack.Push(self)
    self.next_sink.AsyncProcessRequest(sink_stack, msg, buf, headers)

  def AsyncProcessResponse(self, sink_stack, context, stream, msg):
    if msg:
      sink_stack.AsyncProcessResponseMessage(msg)
    else:
      try:
        msg_type, tag = ThriftMuxMessageSerializerSink.ReadHeader(stream)
        msg = self._serializer.Unmarshal(tag, msg_type, stream)
        self._varz.message_bytes_recv(stream.tell())
      except Exception as ex:
        self._varz.deserialization_failures()
        msg = MethodReturnMessage(error=ex)
      sink_stack.AsyncProcessResponseMessage(msg)


ThriftMuxMessageSerializerSink.Builder = SinkProvider(ThriftMuxMessageSerializerSink)


class ClientIdInterceptorSink(ClientMessageSink):
  __slots__ = '_client_id',

  CLIENT_ID_HEADER = 'com.twitter.finagle.thrift.ClientIdContext'

  def __init__(self, next_provider, sink_properties, global_properties):
    super(ClientIdInterceptorSink, self).__init__()
    self._client_id = sink_properties.client_id
    self.next_sink = next_provider.CreateSink(global_properties)

  def AsyncProcessRequest(self, sink_stack, msg, stream, headers):
    msg.properties[self.CLIENT_ID_HEADER] = self._client_id
    self.next_sink.AsyncProcessRequest(sink_stack, msg, stream, headers)

  def AsyncProcessResponse(self, sink_stack, context, stream, msg):
    raise NotImplementedError("This should never be called")


ClientIdInterceptorSink.Builder = SinkProvider(
  ClientIdInterceptorSink,
  client_id='client')



from __future__ import absolute_import

from thrift.protocol.TJSONProtocol import TJSONProtocol, JTYPES, CTYPES
from thrift.Thrift import TType

from ..compat import BytesIO

try:
  import simplejson as json
except ImportError:
  import json


class TFastJSONProtocol(TJSONProtocol):
  class InitContext(object):
    """A context for initializing the reader"""
    def __init__(self, msg):
      self.msg = msg

    def read(self):
      return self.msg

    def write(self, obj):
      self.msg = obj

    def get_buffer(self):
      return self.msg

  class ArrayContext(object):
    """A context for reading from an array."""
    def __init__(self, array):
      assert isinstance(array, list)
      self.arr = array
      self.idx = 0

    def read(self):
      idx = self.idx
      self.idx += 1
      return self.arr[idx]

    def write(self, val):
      self.arr.append(val)

    def get_buffer(self):
      return self.arr

  class ObjectContext(object):
    """A context for reading from an object."""
    def __init__(self, obj):
      assert isinstance(obj, dict)
      self.obj = list(obj.items())
      self.field = None
      self.idx = 0

    def readFieldBegin(self):
      id = 0
      if self.idx >= len(self.obj):
        ttype = TType.STOP
      else:
        obj = self.obj[self.idx]
        id = int(obj[0])
        assert len(obj[1]) == 1
        self.field = obj[1].items()[0]
        ttype = JTYPES[self.field[0]]
      return (None, ttype, id)

    def readFieldEnd(self):
      self.idx += 1

    def writeFieldBegin(self, name, ttype, id):
      self.field = [
        str(id),
        [CTYPES[ttype], None]
      ]

    def writeFieldEnd(self):
      self.field[1] = dict([self.field[1]])
      self.obj.append(self.field)
      self.field = None

    def read(self):
      return self.field[1]

    def write(self, obj):
      self.field[1][1] = obj

    def get_buffer(self):
      return dict(self.obj)

  class MapContext(object):
    """A context for reading from a map."""
    def __init__(self, obj):
      self.map = list(obj.items())
      self.map_idx = 0
      self.idx = 0
      self._read_next()

    def _read_next(self):
      if self.map_idx < len(self.map):
        self.obj = self.map[self.map_idx]
        self.idx = 0
        self.map_idx += 1

    def read(self):
      idx, self.idx = self.idx, self.idx + 1
      o = self.obj[idx]
      if self.idx == 2:
        self._read_next()
      return o

    def write(self, value):
      if self.idx == 0:
        self.map.append([None, None])
      self.map[-1][self.idx] = value
      self.idx += 1
      if self.idx == 2:
        self.idx = 0

    def get_buffer(self):
      return dict(self.map)

  def __init__(self, trans):
    TJSONProtocol.__init__(self, trans)
    self._stack = []
    self._ctx = None

  def _StartReadContext(self, ctx_type):
    next_ctx = ctx_type(self._ctx.read())
    self._stack.append(self._ctx)
    self._ctx = next_ctx

  def _StartWriteContext(self, ctx_type, init):
    next_ctx = ctx_type(init)
    self._stack.append(self._ctx)
    self._ctx = next_ctx

  def readJSONArrayStart(self):
    self._StartReadContext(self.ArrayContext)

  def readJSONObjectStart(self):
    self._StartReadContext(self.ObjectContext)

  def _EndReadContext(self):
    self._ctx = self._stack.pop()

  readMessageEnd = _EndReadContext
  readJSONObjectEnd = _EndReadContext
  readJSONArrayEnd = _EndReadContext

  def _EndWriteContext(self):
    curr = self._ctx.get_buffer()
    self._ctx = self._stack.pop()
    self._ctx.write(curr)

  def _readTransport(self):
    js = BytesIO()
    while True:
      data = self.trans.read(4096)
      if not data:
        break
      js.write(data)
    return js.getvalue()

  def readMessageBegin(self):
    if hasattr(self.trans, 'getvalue'):
      js = self.trans.getvalue()
    else:
      js = self._readTransport()

    message = json.loads(js)
    self._ctx = self.InitContext(message)
    self._stack = []
    return TJSONProtocol.readMessageBegin(self)

  def readJSONString(self, skipContext):
    return self._ctx.read()

  def readJSONInteger(self):
    return int(self._ctx.read())

  def readJSONDouble(self):
    return float(self._ctx.read())

  def readFieldBegin(self):
    return self._ctx.readFieldBegin()

  def readFieldEnd(self):
    return self._ctx.readFieldEnd()

  def readMapBegin(self):
    self.readJSONArrayStart()
    keyType = JTYPES[self.readJSONString(False)]
    valueType = JTYPES[self.readJSONString(False)]
    size = self.readJSONInteger()
    self._StartReadContext(self.MapContext)
    return (keyType, valueType, size)

  def writeMessageBegin(self, name, request_type, seqid):
    self._StartWriteContext(self.InitContext, None)
    TJSONProtocol.writeMessageBegin(self, name, request_type, seqid)

  def writeMessageEnd(self):
    TJSONProtocol.writeMessageEnd(self)
    # The thrift JSON parser is very sensitive, it can't handle spaces after
    # commas or colons <table flip emoji>
    json.dump(self._ctx.get_buffer(), self.trans, separators=(',',':'))

  def writeJSONArrayStart(self):
    self._StartWriteContext(self.ArrayContext, [])

  def writeJSONArrayEnd(self):
    self._EndWriteContext()

  def writeJSONObjectStart(self):
    self._StartWriteContext(self.ObjectContext, {})

  def writeJSONObjectEnd(self):
    self._EndWriteContext()

  def writeFieldBegin(self, name, ttype, id):
    self._ctx.writeFieldBegin(name, ttype, id)

  def writeFieldEnd(self):
    self._ctx.writeFieldEnd()

  def writeMapBegin(self, ktype, vtype, size):
    self.writeJSONArrayStart()
    self.writeJSONString(CTYPES[ktype])
    self.writeJSONString(CTYPES[vtype])
    self.writeJSONNumber(size)
    self._StartWriteContext(self.MapContext, {})

  def writeJSONString(self, number):
    self._ctx.write(number)

  writeJSONNumber = writeJSONString


class TFastJSONProtocolFactory(object):
  def getProtocol(self, trans):
    return TFastJSONProtocol(trans)



from .sink import (
    ClientIdInterceptorSink,
    ThriftMuxMessageSerializerSink,
    SocketTransportSink,
)
from ..core import Scales
from ..loadbalancer import ApertureBalancerSink
from ..resurrector import ResurrectorSink


class ThriftMux(object):
  @staticmethod
  def NewBuilder(Iface, client_id=None):
    builder = Scales.NewBuilder(Iface)
    if client_id:
      builder = builder.WithSink(ClientIdInterceptorSink.Builder(client_id=client_id))
    return builder.WithSink(ThriftMuxMessageSerializerSink.Builder()) \
      .WithSink(ApertureBalancerSink.Builder()) \
      .WithSink(ResurrectorSink.Builder()) \
      .WithSink(SocketTransportSink.Builder())

  @staticmethod
  def NewClient(Iface, uri, timeout=10, client_id=None):
    return ThriftMux.NewBuilder(Iface, client_id=client_id) \
      .SetUri(uri) \
      .SetTimeout(timeout) \
      .Build()



from ..constants import Enum

class MessageType(Enum):
  Tdispatch = 2
  Rdispatch = -2
  Rerr = -128
  BAD_Rerr = 127

  Tping = 65
  Rping = -65

  Tdiscarded = 66
  BAD_Tdiscarded = -62


class Rstatus(Enum):
  OK = 0
  ERROR = 1
  NACK = 2



from thrift.transport.TTransport import TTransportBase

from ..compat import BytesIO
from ..http.sink import HttpTransportSinkBase
from ..sink import SinkProvider


class _ResponseReader(TTransportBase):
  CHUNK_SIZE = 4096

  def __init__(self, response, varz):
    self._stream = response.raw
    self._varz = varz
    self._rbuf = BytesIO()

  def _read_stream(self, sz):
    return self._stream.read(sz, decode_content=True)

  def read(self, sz):
    ret = self._rbuf.read(sz)
    if len(ret) != 0:
      return ret

    data = b''
    while not self._stream.closed:
      data = self._read_stream(max(sz, self.CHUNK_SIZE))
      if data:
        break

    self._varz.bytes_recv(len(data))
    self._rbuf = BytesIO(data)
    return self._rbuf.read(sz)

  def getvalue(self):
    return self._stream.read(decode_content=True)
  # TODO: Implement CReadableTransport


class ThriftHttpTransportSink(HttpTransportSinkBase):
  def __init__(self, next_provider, sink_properties, global_properties):
    self._url_suffix = sink_properties.url
    if self._url_suffix.startswith('/'):
      self._url_suffix = self._url_suffix[1:]
    super(ThriftHttpTransportSink, self).__init__(next_provider, sink_properties, global_properties)

  def _MakeRequest(self, msg, stream, timeout):
    url = 'http://%s:%s/%s' % (
        self._endpoint.host, self._endpoint.port, self._url_suffix)

    content_length = stream.tell()
    headers = {
      'Content-Type': 'application/x-thrift',
      'Content-Length': content_length
    }
    self._varz.bytes_sent(content_length)
    response = self._session.post(
        url,
        data=stream.getvalue(),
        timeout=timeout,
        headers=headers,
        stream=True)
    return response

  def _ProcessResponse(self, response, sink_stack):
    stream = _ResponseReader(response, self._varz)
    sink_stack.AsyncProcessResponseStream(stream)


ThriftHttpTransportSink.Builder = SinkProvider(
  ThriftHttpTransportSink, raise_on_http_error=True, url=None)




import logging
import time

import gevent
from gevent import GreenletExit

from .constants import (ChannelState, SinkProperties)
from .message import (FailedFastError, MethodReturnMessage)
from .sink import (ClientMessageSink, SinkProvider)
from .varz import (AggregateTimer, Counter, Source, VarzBase)

ROOT_LOG = logging.getLogger('scales.Resurrector')


class ResurrectorSink(ClientMessageSink):
  """The resurrector sink monitors its underlying sink for faults, and begins
   attempting to resurrect it.

   Additionally, it will fail any incoming requests immediately while the
   underlying sink is failed.
   """

  class Varz(VarzBase):
    _VARZ_BASE_NAME = 'scales.Resurrector'
    _VARZ = {
      'time_failed': AggregateTimer,
      'reconnect_attempts': Counter,
    }

  def __init__(self, next_factory, sink_properties, global_properties):
    endpoint = global_properties[SinkProperties.Endpoint]
    service = global_properties[SinkProperties.Label]
    endpoint_source = '%s:%d' % (endpoint.host, endpoint.port)
    self.endpoint = endpoint_source
    self._log = ROOT_LOG.getChild('[%s.%s]' % (service, endpoint_source))
    self._properties = global_properties
    self._next_factory = next_factory
    self._resurrector = None
    self._down_on = None
    self._initial_wait_interval = sink_properties.initial_wait_interval
    self._max_wait_interval = sink_properties.max_wait_interval
    self._backoff_exponent = sink_properties.backoff_exponent
    self._varz = self.Varz(Source(service=service,
                                  endpoint=endpoint_source))
    super(ResurrectorSink, self).__init__()

  def AsyncProcessRequest(self, sink_stack, msg, stream, headers):
    if self.next_sink is None:
      # Yield to prevent starvation of the resurrector greenlet if every
      # sink fails.
      gevent.sleep(0)
      sink_stack.AsyncProcessResponseMessage(MethodReturnMessage(error=FailedFastError()))
    else:
      self.next_sink.AsyncProcessRequest(sink_stack, msg, stream, headers)

  def AsyncProcessResponse(self, sink_stack, context, stream, msg):
    raise NotImplementedError("This should never be called")

  def _OnSinkFaulted(self, val):
    if not self._down_on:
      self._down_on = time.time()
      sink, self.next_sink = self.next_sink, None
      sink.Close()
      sink.on_faulted.Unsubscribe(self._OnSinkFaulted)
      self._resurrector = gevent.spawn(self._TryResurrect)
    self.on_faulted.Set(val)

  def _TryResurrect(self):
    if not self._down_on:
      return

    last_attempt = self._down_on
    wait_interval = self._initial_wait_interval
    self._log.info('Attempting to reopen faulted channel')
    while True:
      gevent.sleep(wait_interval)
      now = time.time()
      down_time, last_attempt = now - last_attempt, now
      self._varz.time_failed(down_time)

      sink = self._next_factory.CreateSink(self._properties)
      try:
        self._varz.reconnect_attempts()
        sink.Open().get()
        sink.on_faulted.Subscribe(self._OnSinkFaulted)
        self.next_sink = sink
        self._down_on = None
        self._log.info('Reopened channel.')
        return
      except GreenletExit:
        return
      except:
        sink.Close()

      wait_interval **= self._backoff_exponent
      wait_interval = min(wait_interval, self._max_wait_interval)

  def Open(self):
    if not self.next_sink:
      self.next_sink = self._next_factory.CreateSink(self._properties)
      self.next_sink.on_faulted.Subscribe(self._OnSinkFaulted)
    return self.next_sink.Open()

  def Close(self):
    if self._resurrector:
      self._resurrector.kill(block=False)
      self._resurrector = None
    self._down_on = None
    if self.next_sink:
      self.next_sink.on_faulted.Unsubscribe(self._OnSinkFaulted)
      self.next_sink.Close()

  @property
  def state(self):
    if self._down_on:
      return ChannelState.Closed
    elif not self.next_sink:
      return ChannelState.Idle
    else:
      return self.next_sink.state


ResurrectorSink.Builder = SinkProvider(ResurrectorSink,
  initial_wait_interval=5,
  max_wait_interval=60,
  backoff_exponent=1.2)



import time
from abc import abstractmethod

import gevent
import requests
from requests import exceptions
from six import string_types

from ..asynchronous import AsyncResult
from ..constants import ChannelState, SinkProperties
from ..sink import (ClientMessageSink, SinkProvider)
from ..message import (Deadline, MethodReturnMessage, TimeoutError)
from ..varz import VarzBase, Rate, Source

class HttpTransportSinkBase(ClientMessageSink):
  class Varz(VarzBase):
    _VARZ_BASE_NAME = 'scales.http.HttpTransportSink'
    _VARZ = {
      'bytes_recv': Rate,
      'bytes_sent': Rate,
    }

  def __init__(self, next_provider, sink_properties, global_properties):
    super(HttpTransportSinkBase, self).__init__()
    self._endpoint = global_properties[SinkProperties.Endpoint]
    name = global_properties[SinkProperties.Label]
    self._varz = self.Varz(Source(service=name, endpoint='%s:%d' % (self._endpoint.host, self._endpoint.port)))
    self._open_result = AsyncResult.Complete()
    self._session = requests.Session()
    self._raise_on_http_error = sink_properties.raise_on_http_error

  def Open(self):
    return self._open_result

  def Close(self):
    self._session.close()
    if requests:
      self._session = requests.Session()

  @property
  def state(self):
    return ChannelState.Open

  @abstractmethod
  def _MakeRequest(self, msg, stream, timeout):
    pass

  @abstractmethod
  def _ProcessResponse(self, response, sink_stack):
    pass

  def _DoHttpRequestAsync(self, sink_stack, deadline, stream, msg):
    if deadline:
      timeout = deadline - time.time()
      if timeout < 0:
        sink_stack.AsyncProcessResponseMessage(MethodReturnMessage(error=TimeoutError()))
    else:
      timeout = None

    try:
      response = self._MakeRequest(msg, stream, timeout)
      if self._raise_on_http_error and 400 <= response.status_code < 600:
        err_text = 'HTTP Error %d: %s.' % (response.status_code, response.reason)
        if response.text:
          err_text += '\nThe server returned:\n%s' % response.text
        err = exceptions.HTTPError(err_text, response=response)
        msg = MethodReturnMessage(error=err)
      else:
        self._ProcessResponse(response, sink_stack)
        return
    except exceptions.Timeout:
      msg = MethodReturnMessage(error=TimeoutError())
    except Exception as ex:
      msg = MethodReturnMessage(error=ex)
    sink_stack.AsyncProcessResponseMessage(msg)

  def AsyncProcessRequest(self, sink_stack, msg, stream, headers):
    deadline = msg.properties.get(Deadline.KEY)
    gevent.spawn(
        self._DoHttpRequestAsync,
        sink_stack,
        deadline,
        stream,
        msg)

  def AsyncProcessResponse(self, sink_stack, context, stream, msg):
    pass


class HttpTransportSink(HttpTransportSinkBase):
  def _MakeRequest(self, msg, stream, timeout):
    url, = msg.args
    method = msg.method.lower()
    kwargs = msg.kwargs.copy()

    data = kwargs.get('data')
    if data and isinstance(data, string_types):
      self._varz.bytes_sent(len(data))

    if url.startswith('/'):
      url = url[1:]
    url = 'http://%s:%s/%s' % (self._endpoint.host, self._endpoint.port, url)
    kwargs['timeout'] = timeout
    if method == 'get':
      response = self._session.get(url, **kwargs)
    elif method == 'post':
      response = self._session.post(url, **kwargs)
    elif method == 'put':
      response = self._session.put(url, **kwargs)
    elif method == 'delete':
      response = self._session.delete(url, **kwargs)
    elif method == 'patch':
      response = self._session.patch(url, **kwargs)
    elif method == 'head':
      response = self._session.head(url, **kwargs)
    elif method == 'options':
      response = self._session.options(url, **kwargs)
    else:
      raise NotImplementedError()
    return response

  def _ProcessResponse(self, response, sink_stack):
    self._varz.bytes_recv(len(response.content))
    msg = MethodReturnMessage(response)
    sink_stack.AsyncProcessResponseMessage(msg)

HttpTransportSink.Builder = SinkProvider(
    HttpTransportSink, raise_on_http_error=True)



from .sink import (
  HttpTransportSink
)
from ..core import Scales
from ..loadbalancer.aperture import ApertureBalancerSink
from ..pool import SingletonPoolSink


class _HttpIface(object):
  def Get(self, url, **kwargs):
    """Issue an HTTP GET to url.

    kwargs aligns with arguments to requests, although timeout is ignored.
    """
    pass

  def Post(self, url, **kwargs):
    """Issue an HTTP POST to url.

    kwargs aligns with arguments to requests, although timeout is ignored.
    """
    pass

  def Put(self, url, **kwargs):
    """Issue an HTTP PUT to url.

    kwargs aligns with arguments to requests, although timeout is ignored.
    """
    pass

  def Delete(self, url, **kwargs):
    """Issue an HTTP DELETE to url.

    kwargs aligns with arguments to requests, although timeout is ignored.
    """
    pass

  def Patch(self, url, **kwargs):
    """Issue an HTTP PATCH to url.

    kwargs aligns with arguments to requests, although timeout is ignored.
    """
    pass

  def Head(self, url, **kwargs):
    """Issue an HTTP HEAD to url.

    kwargs aligns with arguments to requests, although timeout is ignored.
    """
    pass

  def Options(self, url, **kwargs):
    """Issue an HTTP OPTIONS to url.

    kwargs aligns with arguments to requests, although timeout is ignored.
    """
    pass


class Http(object):
  @staticmethod
  def NewBuilder():
    return Scales.NewBuilder(_HttpIface)\
      .WithSink(ApertureBalancerSink.Builder())\
      .WithSink(SingletonPoolSink.Builder())\
      .WithSink(HttpTransportSink.Builder())

  @staticmethod
  def NewClient(uri, timeout=60):
    return Http.NewBuilder()\
      .SetUri(uri)\
      .SetTimeout(timeout)\
      .Build()



from struct import (pack, unpack)

from six import string_types

from ..constants import TransportHeaders
from ..message import (
  MethodCallMessage,
  MethodDiscardMessage,
  MethodReturnMessage,
  ServerError,
  Deadline
)
from ..mux.sink import Tag
from ..thrift.serializer import MessageSerializer as ThriftMessageSerializer
from .protocol import (
  Rstatus,
  MessageType
)


class MessageSerializer(object):
  """A serializer that can serialize/deserialize method calls into the ThriftMux
  wire format."""
  def __init__(self, service_cls):
    self._marshal_map = {
      MethodCallMessage: self._Marshal_Tdispatch,
      MethodDiscardMessage: self._Marshal_Tdiscarded,
    }
    self._unmarshal_map = {
      MessageType.Rdispatch: self._Unmarshal_Rdispatch,
      MessageType.Rerr: self._Unmarshal_Rerror,
      MessageType.BAD_Rerr: self._Unmarshal_Rerror,
    }
    if service_cls:
      self._thrift_serializer = ThriftMessageSerializer(service_cls)

  def _Marshal_Tdispatch(self, msg, buf, headers):
    ctx = {}
    ctx.update(msg.public_properties)
    ctx.update(headers)
    MessageSerializer._WriteContext(ctx, buf)

    headers[TransportHeaders.MessageType] = MessageType.Tdispatch
    buf.write(pack('!hh', 0, 0))  # len(dst), len(dtab), both unsupported
    self._thrift_serializer.SerializeThriftCall(msg, buf)

  @staticmethod
  def _Marshal_Tdiscarded(msg, buf, headers):
    headers[TransportHeaders.MessageType] = MessageType.Tdiscarded
    buf.write(pack('!BBB', *Tag(msg.which).Encode()))
    buf.write(msg.reason.encode('utf-8'))

  @staticmethod
  def _WriteContext(ctx, buf):
    buf.write(pack('!h', len(ctx)))
    for k, v in ctx.items():
      if not isinstance(k, string_types):
        raise NotImplementedError("Unsupported key type in context")
      k_len = len(k)
      buf.write(pack('!h%ds' % k_len, k_len, k.encode('utf-8')))
      if isinstance(v, Deadline):
        buf.write(pack('!h', 16))
        buf.write(pack('!qq', v._ts, v._timeout))
      elif isinstance(v, string_types):
        v_len = len(v)
        buf.write(pack('!h%ds' % v_len, v_len, v.encode('utf-8')))
      else:
        raise NotImplementedError("Unsupported value type in context.")

  @staticmethod
  def _ReadContext(buf):
    for _ in range(2):
      sz, = unpack('!h', buf.read(2))
      buf.read(sz)

  def _Unmarshal_Rdispatch(self, buf):
    status, nctx = unpack('!bh', buf.read(3))
    for n in range(0, nctx):
      self._ReadContext(buf)

    if status == Rstatus.OK:
      return self._thrift_serializer.DeserializeThriftCall(buf)
    elif status == Rstatus.NACK:
      return MethodReturnMessage(error=ServerError('The server returned a NACK'))
    else:
      return MethodReturnMessage(error=ServerError(buf.read().decode('utf-8')))

  @staticmethod
  def _Unmarshal_Rerror(buf):
    why = buf.read()
    return MethodReturnMessage(error=ServerError(why.decode('utf-8')))

  def Unmarshal(self, tag, msg_type, buf):
    """Deserialize a message from a stream.

    Args:
      tag - The tag of the message.
      msg_type - The message type intended to be deserialized.
      buf - The stream to deserialize from.
      ctx - The context from serialization.
    Returns:
      A MethodReturnMessage.
    """
    unmarshaller = self._unmarshal_map[msg_type]
    return unmarshaller(buf)

  def Marshal(self, msg, buf, headers):
    """Serialize a message into a stream.

    Args:
      msg - The message to serialize.
      buf - The stream to serialize into.
      headers - (out) Optional headers associated with the message.
    Returns:
      A context to be supplied during deserialization.
    """
    marshaller = self._marshal_map[msg.__class__]
    marshaller(msg, buf, headers)





"""Base classes for Message processing."""

import sys
import traceback

from .compat import Long

class Deadline(object):
  KEY = "__Deadline"
  EVENT_KEY = "__Deadline_Event"

  def __init__(self, timeout):
    """
    Args:
      timeout - The timeout in seconds
    """
    import  time
    self._ts = Long(time.time()) * 1000000000 # Nanoseconds
    self._timeout = Long(timeout * 1000000000)


class ClientError(Exception): pass
class FailedFastError(Exception): pass
class ServerError(Exception): pass
class ChannelConcurrencyError(Exception): pass
class TimeoutError(Exception):
  def __init__(self):
    super(TimeoutError, self).__init__(
      'The call did not complete within the specified timeout '
      'and has been aborted.')


class Message(object):
  """A Message represents the base class for any request."""
  __slots__ = '_properties',

  def __init__(self):
    self._properties = None

  @property
  def is_one_way(self):
    """Returns: True if this method expects no response from the server."""
    return False

  @property
  def properties(self):
    """Returns:
      A dict of properties applying to this message."""
    if not self._properties:
      self._properties = {}
    return self._properties

  @property
  def public_properties(self):
    """Returns:
      A dict of properties intended to be transported to the server
      with the method call."""
    return { k: v for k, v in self.properties.items()
             if not k.startswith('__') }


class MethodCallMessage(Message):
  """A MethodCallMessage represents a method being invoked on a service."""
  __slots__ = ('service', 'method', 'args', 'kwargs')

  def __init__(self, service, method, args, kwargs):
    """
    Args:
      service - The service this method call is intended for.
      method - The method on the service.
      args - The args passed to the method call.
      kwargs-  The kwargs passed to the method call.
    """
    super(MethodCallMessage, self).__init__()
    self.service = service
    self.method = method
    self.args = args
    self.kwargs = kwargs


class MethodDiscardMessage(Message):
  """A MethodDiscardMessage represents the intent of the client to discard the
  response to a message."""
  __slots__ = ('which', 'reason')

  def __init__(self, which, reason):
    """
    Args:
      which - The MethodCallMessage that was discarded.
      reason - The reason the message is being discarded.
    """
    super(MethodDiscardMessage, self).__init__()
    self.which = which
    self.reason = reason

  @property
  def is_one_way(self):
    return True


class MethodReturnMessage(Message):
  """A message representing the return value from a service call."""
  __slots__ = ('return_value', 'error', 'stack')

  def __init__(self, return_value=None, error=None):
    """
    Args:
      return_value - The return value of the call, or None
      error - The error that occurred during processing, or None.
              If not None, the current stack will be captured and included.
    """
    super(MethodReturnMessage, self).__init__()
    self.return_value = return_value
    self.error = error
    if error:
      exc_info = sys.exc_info()
      if len(exc_info) != 3 or exc_info[2] is None:
        try:
          raise ZeroDivisionError
        except ZeroDivisionError:
          tb = sys.exc_info()[2]
          frame = tb.tb_frame.f_back
      else:
        tb = exc_info[2]
        while tb.tb_next is not None:
          tb = tb.tb_next
        frame = tb.tb_frame

      stack = traceback.format_list(traceback.extract_stack(frame))
      stack = stack + traceback.format_exception_only(error.__class__, error)
      self.stack = stack
      # Prevent circular references
      del frame
      del tb
    else:
      self.stack = None



import logging

from abc import (
  abstractmethod
)

from ..message import MethodReturnMessage
from ..sink import ClientMessageSink

ROOT_LOG = logging.getLogger("scales")


class PoolSink(ClientMessageSink):
  """A pool maintains a set of zero or more sinks to a single endpoint."""

  def __init__(self, sink_provider, properties):
    self._properties = properties
    self._sink_provider = sink_provider
    super(PoolSink, self).__init__()

  @abstractmethod
  def _Get(self):
    """To be overridden by subclasses.  Called to get a sink from the pool.

    Returns:
      A sink.
    """
    raise NotImplementedError()

  @abstractmethod
  def _Release(self, sink):
    """To be overridden by subclasses.  Called when a sink has completed
    processing.

    Args:
      sink - A sink that had been previously returned from _Get()
    """
    raise NotImplementedError()

  def AsyncProcessRequest(self, sink_stack, msg, stream, headers):
    """Gets a ClientChannelSink from the pool (asynchronously), and continues
    processing the message on it.
    """
    try:
      sink = self._Get()
    except Exception as e:
      ex_msg = MethodReturnMessage(error=e)
      sink_stack.AsyncProcessResponseMessage(ex_msg)
      return

    sink_stack.Push(self, sink)
    sink.AsyncProcessRequest(sink_stack, msg, stream, headers)

  def AsyncProcessResponse(self, sink_stack, context, stream, msg):
    """Returns the sink (in 'context') to the pool and delegates to the next sink.
    """
    self._Release(context)
    sink_stack.AsyncProcessResponse(stream, msg)



import json
import logging
import posixpath

import gevent
from gevent.event import Event
from gevent.queue import Queue
from kazoo.client import KazooClient
from kazoo.exceptions import NoNodeError
from kazoo.recipe.watchers import (ChildrenWatch, DataWatch)

ROOT_LOG = logging.getLogger('scales.pool.ZooKeeper')

class Endpoint(object):
  """Represents an endpoint in ZooKeeper
  """

  def __init__(self, host, port):
    self._host = host
    self._port = port

  def _key(self):
    return self.host, self.port

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self._key() == other._key()

  def __hash__(self):
    return hash(self._host) ^ hash(self._port)

  @property
  def host(self):
    return self._host

  @property
  def port(self):
    return self._port

  def __str__(self):
    return '%s:%s' % (self.host, self.port)

class Member(object):
  """Represents an instance of a service in ZooKeeper.
  """

  @classmethod
  def from_node(cls, member, data):
    blob = json.loads(data)
    additional_endpoints = blob.get('additionalEndpoints')
    if additional_endpoints is None:
      raise ValueError("Expected additionalEndpoints in member data")
    service_endpoint = blob.get('serviceEndpoint')
    if service_endpoint is None:
      raise ValueError("Expected serviceEndpoint in member data")
    status = blob.get('status')
    if status is None:
      raise ValueError("Expected status in member data")

    shard = blob.get('shard')
    if shard is not None:
      try:
        shard = int(shard)
      except ValueError:
        ROOT_LOG.warn('Unable to parse shard from %r' % shard)
        shard = None

    return cls(
      member=member,
      service_endpoint=Endpoint(service_endpoint['host'], service_endpoint['port']),
      additional_endpoints=dict((name, Endpoint(value['host'], value['port']))
                                for name, value in additional_endpoints.items()),
      shard=shard,
      status=status
    )

  def __init__(
      self,
      member,
      service_endpoint,
      additional_endpoints,
      shard,
      status):

    self._name = member
    self._service_endpoint = service_endpoint
    self._additional_endpoints = additional_endpoints
    self._status = status
    self._shard = shard

  @property
  def name(self):
    return self._name

  @property
  def service_endpoint(self):
    return self._service_endpoint

  @property
  def additional_endpoints(self):
    return self._additional_endpoints

  @property
  def status(self):
    return self._status

  @property
  def shard(self):
    return self._shard

  def __addl_endpoints_str(self):
    return ['%s=>%s' % (k, v) for k, v in self.additional_endpoints.items()]

  def __str__(self):
    return 'Member(%s, %saddl: %s, status: %s)' % (
      self.service_endpoint,
      ('shard: %s, ' % self._shard) if self._shard is not None else '',
      ' : '.join(self.__addl_endpoints_str()),
      self.status
    )

  def _key(self):
    return (
      self.service_endpoint,
      frozenset(sorted(self.__addl_endpoints_str())),
      self.status,
      self._shard)

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self._key() == other._key()

  def __hash__(self):
    return hash(self._key())

class ServerSet(object):
  """A very minimal ServerSet implementation using the Kazoo Client.

  This supports only getting and watching nodes in a ZK ensemble, but is a
  drop in replacement for the twitter.common.zookeeper.ServerSet in those
  cases.
  """

  class _CallbackBlocker(object):
    def __init__(self):
      self.event = Event()
      self.event.set()
      self._count = 0

    def __enter__(self):
      if self._count == 0:
        self.event.clear()
      self._count += 1

    def __exit__(self, exc_type, exc_val, exc_tb):
      self._count -= 1
      if self._count == 0:
        self.event.set()

    def ensure_safe(self):
      self.event.wait()

    def is_blocking(self):
      return self._count != 0

  def __init__(self, zk, zk_path, on_join=None, on_leave=None,
      member_filter=None, member_factory=None):
    """Initialize the ServerSet, ensuring the zk_path exists.

    Args:
      zk - An instance of a Kazoo client.
      zk_path - The path to watch for children in.  If the path does not exist,
                it will be watched for creation.
      on_join - An optional function to call when members join the node.
      on_leave - An optional function to call when members leave the node.
      member_filter - An optional function to filter children from ZK.
      member_factory - An optional function to create a Member object from a znode.
    """
    def noop(*args, **kwargs): pass
    def true(*args, **kwargs): return True
    self._log = ROOT_LOG.getChild('[%s]' % zk_path)
    self._log.info('TellApart ServerSet intializing on path %s' % zk_path)

    if not isinstance(zk, KazooClient):
      raise TypeError('zk must be an instance of a KazooClient')

    if not zk.connected:
      raise Exception('zk must be in a connected state.')

    self._zk_path = zk_path
    self._zk = zk
    self._nodes = set()
    self._members = {}
    self._on_join = on_join or noop
    self._on_leave = on_leave or noop
    self._notification_queue = Queue(0)
    self._watching = False
    self._cb_blocker = self._CallbackBlocker()
    self._member_filter = member_filter or true
    self._member_factory = member_factory or Member.from_node
    self._running = True
    self._worker = gevent.spawn(self._notification_worker)

    if on_join or on_leave:
      self._monitor()

  def stop(self):
    self._running = False
    if self._worker:
      self._worker.kill(block=False)

  def __iter__(self):
    with self._cb_blocker:
      try:
        nodes = self._zk.get_children(self._zk_path)
      except NoNodeError:
        # The un-common case here is if the path doesn't exist,
        # instead of checking every time, assume it exists and catch the exception
        # if it doesn't.
        nodes = ()
      members = self._zk_nodes_to_members(nodes)
      return (n for n in members)

  def get_members(self):
    """Returns a list of members currently in the ServerSet.

    Note: this method makes O(n) calls to ZooKeeper,
    where n = the number of members.
    """
    return list(self)

  def _get_info(self, member):
    """Queries ZooKeeper for the data (service_instance, etc)
    given a member name.

    Args:
      member - The member (relative to zk_path) to get data for.

    Returns:
      The JSON serialized data associated with the node.
    """
    info = self._zk.get(posixpath.join(self._zk_path, member))
    return info[0]

  def _safe_zk_node_to_member(self, node):
    try:
      return self._member_factory(node, self._get_info(node))
    except NoNodeError:
      # Its possible for the ZK node to be removed between getting it
      # from the list and querying it, if so, just skip it.
      return None

  def _zk_nodes_to_members(self, nodes):
    return [m for m in (self._safe_zk_node_to_member(n) for n in nodes
                        if self._member_filter(n))
            if m]

  def _monitor(self):
    """Begins watching the ZK path for node changes.
    """
    if not self._zk.exists(self._zk_path):
      self._log.warn('Path %s does not exist, waiting for it to be created.'
               % self._zk_path)

    # Data changed will notify node on creation / deletion via
    DataWatch(self._zk, self._zk_path, self._data_changed)

  def _data_changed(self, data, stat):
    # stat == None -> the node was deleted (or doesnt exist)
    if stat is None:
      self._watching = False
      self._send_all_removed()
    elif not self._watching:
      self._watching = True
      self._begin_watch()

  def _begin_watch(self):
    self._log.info('Beginning to watch path %s' % self._zk_path)
    ChildrenWatch(self._zk, self._zk_path, self._on_set_changed)

  def _send_all_removed(self):
    for k in self._members.keys():
      member = self._members.pop(k)
      self._on_leave(member)

  def _notification_worker(self):
    """'Atomically' raise notifications for join / leave.

    Having this in a worker prevents multiple updates from interleaving with
    each other, as _zk_nodes_to_members may yield.
    """
    while self._running:
      work = self._notification_queue.get()
      self._cb_blocker.ensure_safe()
      try:
        new_nodes, removed_nodes = work
        new_members = self._zk_nodes_to_members(new_nodes)
        self._members.update(((m.name, m) for m in new_members))

        self._log.debug("Raising notifications for %i members joining and %i members leaving."
                 % (len(new_nodes), len(removed_nodes)))

        for m in removed_nodes:
          removed_member = self._members.pop(m, None)
          if removed_member:
            try:
              self._on_leave(removed_member)
            except Exception:
              self._log.exception('Error in OnLeave callback.')
          else:
            self._log.warn('Member %s was not found in cached set' % str(m))

        for m in new_members:
          try:
            self._on_join(m)
          except Exception:
            self._log.exception('Error in OnJoin callback.')
      except Exception:
        self._log.exception('Error in notification worker.')

  def _on_set_changed(self, children):
    """Called when the children of the watched ZK node change.
    Offloads most work to a greenlet worker thread to do the actual notificaiton.

    Args:
      children - The new set of child nodes.
    """
    children = set([c for c in children if self._member_filter(c)])
    current_nodes = set(self._nodes)
    self._nodes = children
    new_nodes = children - current_nodes
    removed_nodes = current_nodes - children
    self._log.debug("Queueing notifications")
    self._notification_queue.put((new_nodes, removed_nodes))
